import os
import torch

import numpy as np

from abc import abstractmethod
from abc import ABCMeta
from typing import Any
from typing import Dict
from typing import Type
from typing import Tuple
from typing import Union
from typing import TypeVar
from typing import Callable
from typing import Optional
from typing import NamedTuple
from cftool.misc import print_warning
from cftool.misc import Saving
from cftool.misc import WithRegister
from cftool.array import to_device
from cftool.types import tensor_dict_type
from torch.utils.data import Dataset
from torch.utils.data import Sampler
from torch.utils.data import SequentialSampler
from torch.utils.data import DataLoader as TorchDataLoader
from torch.utils.data.distributed import DistributedSampler

from ..types import sample_weights_type
from ..protocol import IDataset
from ..protocol import IDataLoader
from ..misc.toolkit import get_ddp_info
from ..misc.toolkit import get_world_size

try:
    import lmdb
except:
    lmdb = None


data_modules: Dict[str, Type["DataModule"]] = {}
DataModuleType = TypeVar("DataModuleType", bound="DataModule", covariant=True)


class BaseResponse(NamedTuple):
    folder: str
    base: Type["DataModule"]


class LoadInfoResponse(NamedTuple):
    info: Dict[str, Any]
    data_module_bytes: Optional[bytes]


class DataModule(WithRegister[DataModuleType], metaclass=ABCMeta):
    d = data_modules  # type: ignore

    id_file = "id.txt"
    info_name = "info"
    data_folder = "data"
    package_folder = "data_module"

    # inherit

    @property
    @abstractmethod
    def info(self) -> Dict[str, Any]:
        pass

    def prepare(self, sample_weights: sample_weights_type) -> None:
        pass

    def initialize(self) -> Any:
        pass

    # internal

    @abstractmethod
    def _save_data(self, data_folder: str) -> None:
        pass

    @classmethod
    @abstractmethod
    def _load(
        cls,
        data_folder: str,
        info: Dict[str, Any],
        sample_weights: sample_weights_type,
    ) -> "DataModule":
        pass

    def _save_info(self, folder: str) -> None:
        Saving.save_dict(self.info, self.info_name, folder)

    @classmethod
    def _load_info(cls, folder: str) -> Dict[str, Any]:
        return Saving.load_dict(cls.info_name, folder)

    @classmethod
    def _get_base(cls, folder: str) -> BaseResponse:
        folder = os.path.join(folder, cls.package_folder)
        with open(os.path.join(folder, cls.id_file), "r") as f:
            base = cls.get(f.read())
        return BaseResponse(folder, base)

    # api

    def save_info(self, folder: str) -> None:
        """
        Save the core `info` of this `DataModule`

        * The core `info` should be as light-weight as possible.
        * The core `info` represents the minimal subset of information
        which is needed to build up the `Pipeline`, so in most cases this method
        (instead of `save`) will be used to serialize the `DataModule`.
        """
        folder = os.path.join(folder, self.package_folder)
        os.makedirs(folder, exist_ok=True)
        with open(os.path.join(folder, self.id_file), "w") as f:
            f.write(self.__identifier__)
        self._save_info(folder)

    @classmethod
    def load_info(cls, folder: str, *, return_bytes: bool) -> LoadInfoResponse:
        data_module_folder, base = cls._get_base(folder)
        info = base._load_info(data_module_folder)
        if not return_bytes:
            return LoadInfoResponse(info, None)
        Saving.compress(data_module_folder, remove_original=False)
        zip_path = f"{data_module_folder}.zip"
        with open(zip_path, "rb") as f:
            data_module_bytes = f.read()
        os.remove(zip_path)
        return LoadInfoResponse(info, data_module_bytes)

    def save(self, folder: str) -> None:
        """
        Save a complete version of this `DataModule`.

        * This method should serialize every information.
        * Currently this method is only used in the `Experiment` class. See
        `cflearn/dist/ml/experiment` for more details.
        """
        self.save_info(folder)
        data_folder = os.path.join(folder, self.package_folder, self.data_folder)
        os.makedirs(data_folder, exist_ok=True)
        self._save_data(data_folder)

    @classmethod
    def load(
        cls,
        folder: str,
        *,
        sample_weights: sample_weights_type = None,
    ) -> "DataModule":
        folder, base = cls._get_base(folder)
        info = base._load_info(folder)
        data_folder = os.path.join(folder, cls.data_folder)
        return base._load(data_folder, info, sample_weights)


data_loaders_type = Tuple[IDataLoader, Optional[IDataLoader]]


@DataModule.register("dl")  # type: ignore
class DLDataModule(DataModule, metaclass=ABCMeta):
    train_data: IDataset
    valid_data: Optional[IDataset]

    @abstractmethod
    def initialize(self) -> data_loaders_type:
        pass

    def get_loaders(
        self,
        *,
        sample_weights: sample_weights_type = None,
    ) -> data_loaders_type:
        self.prepare(sample_weights)
        return self.initialize()


class DLDataset(IDataset):
    def __init__(self, dataset: Dataset):
        super().__init__()
        self.dataset = dataset

    def __len__(self) -> int:
        return len(self.dataset)  # type: ignore

    def __getitem__(self, item: Any) -> Any:
        return self.dataset[item]


class DataLoader(TorchDataLoader):
    def __init__(
        self,
        dataset: Dataset,
        batch_size: Optional[int] = 1,
        shuffle: bool = False,
        sampler: Optional[Sampler[int]] = None,
        *args: Any,
        use_distributed_sampler: Optional[bool] = None,
        **kwargs: Any,
    ):
        if use_distributed_sampler is None:
            use_distributed_sampler = get_ddp_info() is not None
        if use_distributed_sampler:
            if sampler is not None and not isinstance(sampler, DistributedSampler):
                raise ValueError(
                    "`sampler` should be `DistributedSampler` "
                    "when `use_distributed_sampler` is True"
                )
            sampler = DistributedSampler(dataset, shuffle=shuffle)
            shuffle = False
        super().__init__(dataset, batch_size, shuffle, sampler, *args, **kwargs)

    def __setattr__(self, attr: str, val: Any) -> None:
        if self.__initialized and attr in (
            "batch_size",
            "batch_sampler",
            "drop_last",
            "dataset",
            "persistent_workers",
        ):
            raise ValueError(
                f"{attr} attribute should not be set after "
                f"{self.__class__.__name__} is initialized"
            )

        super(TorchDataLoader, self).__setattr__(attr, val)


class DLLoader(IDataLoader):
    data: DLDataset
    next_batch: tensor_dict_type

    def __init__(
        self,
        loader: DataLoader,
        batch_callback: Optional[Callable[[Any], tensor_dict_type]] = None,
        *,
        sample_weights: Optional[np.ndarray] = None,
        prefetch_device: Optional[Union[int, str]] = None,
    ):
        if sample_weights is not None:
            raise ValueError(
                "in `DLLoader`, we should introduce `sample_weights` to the original "
                "Pytorch `DataLoader` (by specifying corresponding samplers)"
            )
        super().__init__(sample_weights=sample_weights)
        self.loader = loader
        self.data = loader.dataset  # type: ignore
        self.batch_callback = batch_callback
        self.sampler_backup = loader.sampler
        self._iterator: Optional[Any] = None
        # prefetch stuffs
        if prefetch_device is not None and torch.cuda.is_available():
            prefetch_device = f"cuda:{prefetch_device}"
            self.stream = torch.cuda.Stream(prefetch_device)
        else:
            if prefetch_device is not None:
                print_warning(
                    "`prefetch_device` is specified but "
                    "cuda is not available, it will have no effects"
                )
            self.stream = None
        self.device = prefetch_device or torch.device("cpu")
        self.stop_at_next_batch = False

    def __iter__(self) -> "DLLoader":
        self.stop_at_next_batch = False
        self._iterator = self.loader.__iter__()
        self.preload()
        return self

    def __next__(self) -> tensor_dict_type:
        if self.stop_at_next_batch:
            raise StopIteration
        if self.stream is not None:
            torch.cuda.current_stream(self.device).wait_stream(self.stream)
        batch = self.next_batch
        self.preload()
        return batch

    def __len__(self) -> int:
        return len(self.loader)

    def preload(self) -> None:
        try:
            batch = self._iterator.__next__()  # type: ignore
        except StopIteration:
            self.stop_at_next_batch = True
            return None

        if self.batch_callback is not None:
            batch = self.batch_callback(batch)
        if self.stream is not None:
            with torch.cuda.stream(self.stream):
                batch = to_device(batch, self.device, non_blocking=True)
        batch = to_device(batch, self.device, non_blocking=True)
        self.next_batch = batch

    @property
    def batch_size(self) -> int:  # type: ignore
        return self.loader.batch_size * get_world_size()

    def copy(self) -> "DLLoader":
        stream = self.stream
        dataset = self.data.dataset
        self.__dict__.pop("stream")
        self.data.__dict__.pop("dataset")
        copied = super().copy()
        assert isinstance(copied, DLLoader)
        self.stream = copied.stream = stream
        self.data.dataset = copied.data.dataset = dataset
        return copied

    def disable_shuffle(self) -> None:
        sampler = SequentialSampler(self.data)
        self.loader.sampler = sampler
        if hasattr(self.loader, "batch_sampler"):
            self.loader.batch_sampler.sampler = sampler

    def recover_shuffle(self) -> None:
        self.loader.sampler = self.sampler_backup
        if hasattr(self.loader, "batch_sampler"):
            self.loader.batch_sampler.sampler = self.sampler_backup


__all__ = [
    "DataModule",
    "DLDataModule",
    "DLDataset",
    "DataLoader",
    "DLLoader",
]



import torch

from torch import nn
from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ....constants import LATENT_KEY
from ..encoder.backbone import BackboneEncoder
from ...protocols.cv import ImageTranslatorMixin
from ....misc.toolkit import interpolate
from ....modules.blocks import get_conv_blocks
from ....modules.blocks import Conv2d
from ....modules.blocks import Lambda
from ....modules.blocks import Linear
from ....misc.internal_.register import register_module


@register_module("linear_seg")
class LinearSegmentation(nn.Module, ImageTranslatorMixin):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        latent_dim: int = 768,
        *,
        dropout: float = 0.1,
        backbone: str = "mix_vit",
        backbone_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.backbone = BackboneEncoder(
            backbone,
            in_channels,
            backbone_config=backbone_config,
        )
        linear_blocks: List[nn.Module] = []
        backbone_channels = self.backbone.net.out_channels
        for num_channel in backbone_channels:
            linear_blocks.append(
                nn.Sequential(
                    Lambda(lambda t: t.flatten(2).transpose(1, 2), "BCHW -> BNC"),
                    Linear(num_channel, latent_dim),
                )
            )
        self.linear_blocks = nn.ModuleList(linear_blocks)
        self.linear_fuse = nn.Sequential(
            *get_conv_blocks(
                latent_dim * 4,
                latent_dim,
                kernel_size=1,
                stride=1,
                norm_type="batch",
            )
        )
        self.dropout = nn.Dropout2d(dropout) if dropout > 0.0 else None
        self.linear_head = Conv2d(latent_dim, out_channels, kernel_size=1)

    def forward(self, inp: Tensor) -> Tensor:
        features = self.backbone(inp)
        features.pop(LATENT_KEY)
        outputs = []
        for i, linear in enumerate(self.linear_blocks):
            net = features[f"stage{i + 1}"]
            b, _, h, w = net.shape
            net = linear(net).transpose(1, 2)
            net = net.contiguous().view(b, -1, h, w)
            net = interpolate(net, anchor=inp, mode="bilinear")
            outputs.append(net)
        net = torch.cat(outputs, dim=1)
        net = self.linear_fuse(net)
        if self.dropout is not None:
            net = self.dropout(net)
        net = self.linear_head(net)
        return net


__all__ = ["LinearSegmentation"]



import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from .protocol import VanillaGANMixin
from ..decoder import make_decoder
from ....constants import INPUT_KEY
from ....constants import LABEL_KEY
from ....misc.internal_.register import register_custom_module
from ....misc.internal_.register import CustomModule


@register_custom_module("gan")
class VanillaGAN(VanillaGANMixin, CustomModule):  # type: ignore
    generator: nn.Module

    def __init__(
        self,
        img_size: int,
        in_channels: int,
        out_channels: Optional[int] = None,
        latent_dim: int = 64,
        latent_resolution: int = 7,
        *,
        generator: str = "vanilla",
        discriminator: str = "basic",
        generator_config: Optional[Dict[str, Any]] = None,
        discriminator_config: Optional[Dict[str, Any]] = None,
        num_classes: Optional[int] = None,
        gan_mode: str = "vanilla",
        gan_loss_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self._initialize(
            in_channels=in_channels,
            discriminator=discriminator,
            discriminator_config=discriminator_config,
            num_classes=num_classes,
            gan_mode=gan_mode,
            gan_loss_config=gan_loss_config,
        )
        self.latent_dim = latent_dim
        # generator
        if generator_config is None:
            generator_config = {}
        generator_config["img_size"] = img_size
        generator_config["latent_dim"] = latent_dim
        generator_config["latent_resolution"] = latent_resolution
        generator_config["out_channels"] = out_channels or in_channels
        generator_config["num_classes"] = num_classes
        self.generator = make_decoder(generator, generator_config, is_1d=True)

    @property
    def g_parameters(self) -> List[nn.Parameter]:
        return list(self.generator.parameters())

    def decode(self, z: Tensor, *, labels: Optional[Tensor], **kwargs: Any) -> Tensor:
        batch = {INPUT_KEY: z, LABEL_KEY: labels}
        net = self.generator.decode(batch, **kwargs)
        return net


__all__ = ["VanillaGAN"]



import math
import torch

import numpy as np
import torch.nn as nn
import torch.nn.functional as F

from typing import Any
from typing import Dict
from typing import List
from typing import Union
from typing import Optional
from typing import NamedTuple
from collections import defaultdict
from cftool.array import arr_type
from cftool.array import to_torch

from ...data import MLLoader
from ...constants import INPUT_KEY
from ...misc.toolkit import Initializer
from ...modules.blocks import Lambda


class EncodingResult(NamedTuple):
    one_hot: Optional[torch.Tensor]
    embedding: Optional[torch.Tensor]

    @property
    def merged(self) -> torch.Tensor:
        if self.one_hot is None and self.embedding is None:
            raise ValueError("no data is provided in `EncodingResult`")
        if self.one_hot is None:
            assert self.embedding is not None
            return self.embedding
        if self.embedding is None:
            assert self.one_hot is not None
            return self.one_hot
        return torch.cat([self.one_hot, self.embedding], dim=-1)


class OneHot(nn.Module):
    def __init__(self, dim: int):
        super().__init__()
        one_hot_fn = lambda column: F.one_hot(column, dim)
        self.core = Lambda(one_hot_fn, f"one_hot_{dim}")
        self.dim = dim

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return self.core(tensor).to(torch.float32)


class Embedding(nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        init_method: Optional[str],
        init_config: Dict[str, Any],
    ):
        super().__init__()
        weights = torch.empty(in_dim, out_dim)
        if init_method is None:
            nn.init.normal_(weights)
        else:
            Initializer(init_config).initialize(weights, init_method)
        self.weights = nn.Parameter(weights)
        self.in_dim, self.out_dim = in_dim, out_dim

    def extra_repr(self) -> str:
        return f"{self.in_dim} -> {self.out_dim}"

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return F.embedding(tensor, self.weights)


class EncodingSettings(NamedTuple):
    """
    Encoder settings.

    Properties
    ----------
    dim (int) : number of different values of this categorical column.
    methods (str | List[str]) : encoding methods to use for each categorical column.
        * if List[str] is provided and its length > 1, then multiple encoding methods will be used.
    method_configs (Dict[str, Any]) : (flattened) configs of the corresponding encoding methods.
        * even if multiple methods are used, `method_configs` should still be 'flattened'

    """

    dim: int
    methods: Union[str, List[str]] = "embedding"
    method_configs: Optional[Dict[str, Any]] = None

    @property
    def use_one_hot(self) -> bool:
        if self.methods == "one_hot":
            return True
        if isinstance(self.methods, list) and "one_hot" in self.methods:
            return True
        return False

    @property
    def use_embedding(self) -> bool:
        if self.methods == "embedding":
            return True
        if isinstance(self.methods, list) and "embedding" in self.methods:
            return True
        return False


class Encoder(nn.Module):
    def __init__(
        self,
        settings: Dict[int, EncodingSettings],
        *,
        config: Optional[Dict[str, Any]] = None,
        loaders: Optional[List[MLLoader]] = None,
    ):
        """
        Encoder to perform differentiable categorical encodings.

        Terminology
        ----------
        `i`   : indicates the {i}th element of `columns`.
        `idx` : indicates the actual column index.

        -> So basically, we have self.columns[i] = idx.

        Parameters
        ----------
        settings (Dict[`idx`, EncoderSettings]) : encoding settings of each categorical column.
            * key: `idx`.
            * value: `EncoderSettings`.
        config (Dict[str, Any]) : configs of the encoder.
            * embedding_dropout (float) : dropout rate for embedding.
                * default : 0.2
            * default_embedding_init_method (str) : default init method for embedding.
                * default : "truncated_normal"
            * default_embedding_init_config (Dict[str, Any]) : default init config for embedding.
                * default : {"mean": 0, "std": 0.02}
            * use_fast_embedding (bool) : whether to use fast embedding.
                * default : True
            * recover_original_dim (bool) : whether to recover original dim when fast embedding is applied.
                * default : False
            * unified_embedding_dim (str | int) : unified embedding dim used in fast embedding.
                * default : "max"
            * fast_embedding_init_method (str | None) : init method for fast embedding.
                * default : None
            * fast_embedding_init_config (Dict[str, Any] | None) : init config for fast embedding.
                * default : None
        loaders (List[MLLoader]) : internally used by `carefree-learn`.

        Properties
        ----------
        columns (List[`idx`]) : sorted categorical columns.
        input_dims (tensor) : number of different values of each categorical column.
            * len(input_dims) = len(columns)
            * input_dims[i] indicates the number of different values of the {columns[i]}th column.
        _one_hot_indices (List[`i`]) : indices that are encoded with one-hot.
        _one_hot_columns (List[`idx`]) : columns that are encoded with one-hot.
        _embed_indices (List[`i`]) : indices that are encoded with embedding.
        _embed_columns (List[`idx`]) : indices that are encoded with embedding.
        _embed_dims (Dict[`idx`, int]) : dims of embedding used for each categorical column.
            * key : `idx`.
            * value : embed dimension of the embedding module.
            * len(_embed_dims) should be = len(_embed_indices).
        merged_dims (Dict[`idx`, int])
            * key : `idx`.
            * value : final encoded dim of the {key}th column.
        embeddings (nn.ModuleDict) : collection of embedding modules.
            * key : str(`idx`).
            * value : embedding module of the {key}th column.
        one_hot_encoders (nn.ModuleDict) : collection of one-hot encoder modules.
            * key : str(`idx`).
            * value : one-hot encoder module of the {key}th column.

        """
        super().__init__()
        loaders = loaders or []
        self._fe_init_method: Optional[str]
        self._fe_init_config: Optional[Dict[str, Any]]
        self._init_config(config or {})
        for loader in loaders:
            if loader.shuffle:
                raise ValueError("`loader` should not be shuffled in `Encoder`")
        self.merged_dim = 0
        self.one_hot_dim = 0
        self.embedding_dim = 0
        self.columns = sorted(settings)
        self.tgt_columns = np.array(self.columns, np.int64)
        self.register_buffer(
            "input_dims",
            torch.tensor(
                [settings[idx].dim for idx in self.columns],
                dtype=torch.float32,
            ),
        )
        self.merged_dims: Dict[int, int] = defaultdict(int)
        self.embeddings = nn.ModuleDict()
        self.one_hot_encoders = nn.ModuleDict()
        self._one_hot_indices: List[int] = []
        self._one_hot_columns: List[int] = []
        self._embed_indices: List[int] = []
        self._embed_columns: List[int] = []
        self._embed_dims: Dict[int, int] = {}
        for i, idx in enumerate(self.columns):
            self._register(i, idx, settings[idx])
        # fast embedding
        self.unified_dim = 0
        if self.use_embedding and self._use_fast_embed:
            sorted_dims = sorted(self._embed_dims.values())
            max_dim = sorted_dims[-1]
            if isinstance(self._unified_embed_dim, int):
                unified_dim = self._unified_embed_dim
            else:
                if self._unified_embed_dim == "max":
                    unified_dim = max_dim
                elif self._unified_embed_dim == "mean":
                    unified_dim = int(round(sum(sorted_dims) / self.num_embedding))
                elif self._unified_embed_dim == "median":
                    half_idx = self.num_embedding // 2
                    if self.num_embedding % 2 != 0:
                        unified_dim = sorted_dims[half_idx]
                    else:
                        left = sorted_dims[half_idx - 1]
                        right = sorted_dims[half_idx]
                        unified_dim = int(round(0.5 * (left + right)))
                else:
                    raise NotImplementedError(
                        f"unified embedding dim '{self._unified_embed_dim}' "
                        "is not recognized"
                    )
            if self._recover_dim and unified_dim < max_dim:
                raise ValueError(
                    f"unified embedding dim {unified_dim} is smaller than "
                    f"max embedding dim {max_dim}, which is not allowed "
                    f"when `_recover_dim` is set to True"
                )
            self.unified_dim = unified_dim
            embedding_dim_sum = 0
            for idx in self._embed_columns:
                original_dim = self._embed_dims[idx]
                final_dim = original_dim if self._recover_dim else unified_dim
                self.merged_dims[idx] += final_dim
                embedding_dim_sum += final_dim
            self.embedding_dim = embedding_dim_sum
            self.merged_dim += embedding_dim_sum
            if self._fe_init_method is None:
                self._fe_init_method = self._de_init_method
            if self._fe_init_config is None:
                self._fe_init_config = self._de_init_config
            assert isinstance(self._fe_init_config, dict)
            embed_input_dims = [settings[idx].dim for idx in self._embed_columns]
            self.embeddings["-1"] = Embedding(
                sum(embed_input_dims),
                unified_dim,
                self._fe_init_method,
                self._fe_init_config,
            )
            embed_input_dims_cumsum = torch.tensor(embed_input_dims).cumsum(0)[:-1]
            embed_input_dims_cumsum = embed_input_dims_cumsum.to(torch.long)
            self.register_buffer("embed_input_dims_cumsum", embed_input_dims_cumsum)
            if not self._recover_dim or len(set(self._embed_dims.values())) == 1:
                self.register_buffer("recover_indices", None)
            else:
                recover_indices: List[int] = []
                for i, idx in enumerate(self._embed_columns):
                    dim = self._embed_dims[idx]
                    recover_indices.extend(i * unified_dim + j for j in range(dim))
                recover_indices_tensor = torch.tensor(recover_indices, dtype=torch.long)
                self.register_buffer("recover_indices", recover_indices_tensor)
        # embedding dropout
        self.embedding_dropout = None
        if self.use_embedding and 0.0 < self._embed_drop < 1.0:
            self.embedding_dropout = nn.Dropout(self._embed_drop)
        # compile
        self._all_one_hot = len(self._one_hot_indices) == len(settings)
        self._all_embedding = len(self._embed_indices) == len(settings)
        self._compile(loaders)

    @property
    def num_one_hot(self) -> int:
        return len(self._one_hot_indices)

    @property
    def num_embedding(self) -> int:
        return len(self._embed_indices)

    @property
    def use_one_hot(self) -> bool:
        return self.num_one_hot > 0

    @property
    def use_embedding(self) -> bool:
        return self.num_embedding > 0

    def forward(
        self,
        x_batch: torch.Tensor,
        batch_indices: Optional[np.ndarray],
        loader_name: Optional[str],
    ) -> EncodingResult:
        keys = None
        if loader_name is not None:
            keys = self._get_cache_keys(loader_name)
        categorical_columns = x_batch[..., self.tgt_columns]
        if batch_indices is None or loader_name is None:
            categorical_columns = self._oob_imputation(categorical_columns)
        use_cache = keys is not None and batch_indices is not None
        # one hot
        if not self.use_one_hot:
            one_hot = None
        else:
            if use_cache:
                one_hot = getattr(self, keys["one_hot"])[batch_indices]  # type: ignore
            else:
                one_hot_columns = categorical_columns
                if not self._all_one_hot:
                    one_hot_columns = one_hot_columns[..., self._one_hot_indices]
                one_hot = self._one_hot(one_hot_columns)
        # embedding
        if not self.use_embedding:
            embedding = None
        else:
            use_cache = use_cache and self._use_fast_embed
            if not use_cache:
                indices = categorical_columns
            else:
                indices = getattr(self, keys["indices"])[batch_indices]  # type: ignore
            if not self._all_embedding:
                indices = indices[..., self._embed_indices]
            if not use_cache and self._use_fast_embed:
                indices[..., 1:] += self.embed_input_dims_cumsum
            embedding = self._embedding(indices.to(torch.long))
            if self.embedding_dropout is not None:
                embedding = self.embedding_dropout(embedding)
        return EncodingResult(one_hot, embedding)

    def encode(self, column: arr_type, column_idx: int) -> EncodingResult:
        if column_idx not in self.columns:
            return EncodingResult(None, None)
        if isinstance(column, np.ndarray):
            column = to_torch(column)
        column = column.to(torch.long).view(-1)
        # one hot
        if not self.use_one_hot or column_idx not in self._one_hot_columns:
            one_hot = None
        else:
            one_hot = self.one_hot_encoders[str(column_idx)](column)
        # embedding
        if not self.use_embedding or column_idx not in self._embed_columns:
            embedding = None
        else:
            indices = column.clone()
            if not self._use_fast_embed:
                embedding = self.embeddings[str(column_idx)](indices)
            else:
                i_embedding = self._embed_columns.index(column_idx)
                if i_embedding > 0:
                    indices += self.embed_input_dims_cumsum[i_embedding - 1]
                embedding = self.embeddings["-1"](indices)
                if self._recover_dim:
                    dim = self._embed_dims[column_idx]
                    embedding = embedding[..., list(range(dim))]
            if self.embedding_dropout is not None:
                embedding = self.embedding_dropout(embedding)
        return EncodingResult(one_hot, embedding)

    def _init_config(self, config: Dict[str, Any]) -> None:
        self.config = config
        self._embed_drop = config.setdefault("embedding_dropout", 0.2)
        self._de_init_method = config.setdefault(
            "default_embedding_init_method",
            "truncated_normal",
        )
        self._de_init_config = config.setdefault(
            "default_embedding_init_config",
            {"mean": 0.0, "std": 0.02},
        )
        self._use_fast_embed = config.setdefault("use_fast_embedding", True)
        self._recover_dim = config.setdefault("recover_original_dim", False)
        # [ mean | median | max | int ]
        self._unified_embed_dim = config.setdefault("unified_embedding_dim", "max")
        self._fe_init_method = config.setdefault("fast_embedding_init_method", None)
        self._fe_init_config = config.setdefault("fast_embedding_init_config", None)

    def _register(self, i: int, idx: int, setting: EncodingSettings) -> None:
        methods = setting.methods
        if isinstance(methods, str):
            methods = [methods]
        for method in methods:
            attr = getattr(self, f"_register_{method}", None)
            if attr is None:
                msg = f"encoding method '{method}' is not implemented"
                raise NotImplementedError(msg)
            attr(i, idx, setting.dim, setting.method_configs or {})

    def _register_one_hot(
        self,
        i: int,
        idx: int,
        in_dim: int,
        _: Dict[str, Any],
    ) -> None:
        self.one_hot_encoders[str(idx)] = OneHot(in_dim)
        self._one_hot_indices.append(i)
        self._one_hot_columns.append(idx)
        self.merged_dims[idx] += in_dim
        self.one_hot_dim += in_dim
        self.merged_dim += in_dim

    @staticmethod
    def _get_embed_key(num: int) -> str:
        return f"embedding_weight_{num}"

    def _register_embedding(
        self,
        i: int,
        idx: int,
        in_dim: int,
        config: Dict[str, Any],
    ) -> None:
        self._embed_indices.append(i)
        self._embed_columns.append(idx)
        embedding_dim = config.setdefault("embedding_dim", "auto")
        if isinstance(embedding_dim, int):
            out_dim = embedding_dim
        elif embedding_dim == "log":
            out_dim = math.ceil(math.log2(in_dim))
        elif embedding_dim == "sqrt":
            out_dim = math.ceil(math.sqrt(in_dim))
        elif embedding_dim == "auto":
            out_dim = min(in_dim, max(4, min(8, math.ceil(math.log2(in_dim)))))
        else:
            raise ValueError(f"embedding dim '{embedding_dim}' is not defined")
        if self._use_fast_embed:
            self._embed_dims[idx] = out_dim
            if self._fe_init_method is None:
                self._fe_init_method = config.get("init_method")
            if self._fe_init_config is None:
                self._fe_init_config = config.get("init_config")
        else:
            init_method = config.setdefault("init_method", self._de_init_method)
            init_config = config.setdefault("init_config", self._de_init_config)
            embedding = Embedding(in_dim, out_dim, init_method, init_config)
            self.embeddings[str(idx)] = embedding
            self.merged_dims[idx] += out_dim
            self.embedding_dim += out_dim
            self.merged_dim += out_dim

    def _oob_imputation(
        self,
        categorical_columns: torch.Tensor,
        batch_indices: Optional[np.ndarray] = None,
        loader_name: Optional[str] = None,
    ) -> torch.Tensor:
        if loader_name is None or batch_indices is None:
            oob_mask = categorical_columns >= self.input_dims
        else:
            keys = self._get_cache_keys(loader_name)
            oob_mask = getattr(self, keys["oob"])[batch_indices].to(torch.bool)
        return torch.where(
            oob_mask,
            torch.zeros_like(categorical_columns),
            categorical_columns,
        )

    @staticmethod
    def _to_split(columns: torch.Tensor) -> List[torch.Tensor]:
        return list(columns.to(torch.long).unbind(dim=-1))

    def _one_hot(self, one_hot_columns: torch.Tensor) -> torch.Tensor:
        split = self._to_split(one_hot_columns)
        encodings = [
            self.one_hot_encoders[str(self._one_hot_columns[i])](flat_feature)
            for i, flat_feature in enumerate(split)
        ]
        return torch.cat(encodings, dim=-1)

    def _embedding(self, indices_columns: torch.Tensor) -> torch.Tensor:
        if self._use_fast_embed:
            embed_mat = self.embeddings["-1"](indices_columns)
            embed_mat = embed_mat.view(
                *indices_columns.shape[:-1],
                self.num_embedding * self.unified_dim,
            )
            if not self._recover_dim or self.recover_indices is None:
                return embed_mat
            return embed_mat[..., self.recover_indices]
        split = self._to_split(indices_columns)
        encodings = [
            self.embeddings[str(self._embed_columns[i])](flat_feature)
            for i, flat_feature in enumerate(split)
        ]
        return torch.cat(encodings, dim=-1)

    @staticmethod
    def _get_cache_keys(name: str) -> Dict[str, str]:
        return {
            "one_hot": f"{name}_one_hot_cache",
            "indices": f"{name}_indices_cache",
            "oob": f"{name}_oob_cache",
        }

    def _compile(self, loaders: List[MLLoader]) -> None:
        for loader in loaders:
            name = loader.name
            if name is None:
                continue
            categorical_features = []
            for sample in loader:
                x_batch = sample[INPUT_KEY]
                categorical_features.append(x_batch[..., self.tgt_columns])
            tensor = to_torch(np.vstack(categorical_features))
            keys = self._get_cache_keys(name)
            # compile oob
            oob = (tensor >= self.input_dims).to(torch.float32)
            self.register_buffer(keys["oob"], oob)
            tensor = self._oob_imputation(
                tensor,
                batch_indices=np.arange(len(tensor)),
                loader_name=name,
            )
            # compile one hot
            if self.use_one_hot:
                one_hot_cache = self._one_hot(tensor[..., self._one_hot_indices])
                self.register_buffer(keys["one_hot"], one_hot_cache)
            # compile embedding
            if self.use_embedding and self._use_fast_embed:
                tensor[..., 1:] += self.embed_input_dims_cumsum
                indices = tensor.to(torch.long)
                self.register_buffer(keys["indices"], indices.to(torch.float32))


__all__ = [
    "Encoder",
    "EncodingSettings",
    "EncodingResult",
]



import torch

from typing import List
from typing import Optional
from cftool.types import tensor_dict_type

from .fcnn import FCNN
from .linear import Linear
from ..register import register_ml_module
from ..protocols.ml import ONE_HOT_KEY
from ..protocols.ml import EMBEDDING_KEY
from ..protocols.ml import NUMERICAL_KEY
from ..protocols.ml import Dimensions


@register_ml_module("wnd")
class WideAndDeep(torch.nn.Module):
    def __init__(
        self,
        output_dim: int,
        num_history: int,
        dimensions: Dimensions,
        hidden_units: Optional[List[int]] = None,
        *,
        mapping_type: str = "basic",
        bias: bool = True,
        activation: str = "ReLU",
        batch_norm: bool = False,
        dropout: float = 0.0,
    ):
        super().__init__()
        wide_dim = dimensions.categorical_dim or dimensions.numerical_dim
        deep_dim = dimensions.numerical_dim + dimensions.embedding_dim
        self.fcnn = FCNN(
            deep_dim,
            output_dim,
            num_history,
            hidden_units,
            mapping_type=mapping_type,
            bias=bias,
            activation=activation,
            batch_norm=batch_norm,
            dropout=dropout,
        )
        self.linear = Linear(wide_dim, output_dim, num_history, bias=bias)

    def forward(self, batch: tensor_dict_type) -> tensor_dict_type:
        one_hot = batch[ONE_HOT_KEY]
        embedding = batch[EMBEDDING_KEY]
        numerical = batch[NUMERICAL_KEY]
        # wide
        if one_hot is None and embedding is None:
            wide_inp = numerical
        else:
            if one_hot is None:
                wide_inp = embedding
            elif embedding is None:
                wide_inp = one_hot
            else:
                wide_inp = torch.cat([one_hot, embedding], dim=-1)
        wide_net = self.linear(wide_inp)
        # deep
        if embedding is None:
            deep_inp = numerical
        elif numerical is None:
            deep_inp = embedding
        else:
            deep_inp = torch.cat([numerical, embedding], dim=-1)
        deep_net = self.fcnn(deep_inp)
        # merge
        return wide_net + deep_net


__all__ = ["WideAndDeep"]



from typing import Any
from typing import Tuple
from typing import Union
from typing import Optional

from .general import ToRGB
from .general import ToGray
from .general import BatchWrapper
from ..basic import Transforms
from ....constants import INPUT_KEY
from ....constants import LABEL_KEY

try:
    import cv2
except:
    cv2 = None
try:
    import albumentations as A
    from albumentations.pytorch import ToTensorV2
except:
    A = ToTensorV2 = None


class ATransforms(Transforms):
    input_alias = "image"

    def __init__(self, *, label_alias: Optional[str] = None):
        super().__init__()
        self.label_alias = label_alias
        if A is None:
            msg = f"`albumentations` is needed for `{self.__class__.__name__}`"
            raise ValueError(msg)

    def __call__(self, inp: Any, **kwargs: Any) -> Any:  # type: ignore
        if not self.need_batch_process:
            kwargs[self.input_alias] = inp
            return self.fn(**kwargs)[self.input_alias]
        inp_keys_mapping = {
            self.input_alias
            if k == INPUT_KEY
            else self.label_alias
            if k == LABEL_KEY
            else k: k
            for k in inp
        }
        inp = {k: inp[v] for k, v in inp_keys_mapping.items()}
        return {inp_keys_mapping[k]: v for k, v in self.fn(**inp).items()}

    @property
    def need_batch_process(self) -> bool:
        return self.label_alias is not None

    @property
    def need_numpy(self) -> bool:
        return True


AToRGB = lambda: BatchWrapper(ToRGB(), ATransforms.input_alias)
AToGray = lambda: BatchWrapper(ToGray(), ATransforms.input_alias)


@Transforms.register("a_resize")
class AResize(ATransforms):
    def __init__(
        self,
        size: Union[int, tuple] = 224,
        *,
        label_alias: Optional[str] = None,
    ):
        super().__init__(label_alias=label_alias)
        if isinstance(size, int):
            size = size, size
        self.fn = A.Resize(*size)


@Transforms.register("a_random_crop")
class ARandomCrop(ATransforms):
    def __init__(self, size: Union[int, tuple], *, label_alias: Optional[str] = None):
        super().__init__(label_alias=label_alias)
        if isinstance(size, int):
            size = size, size
        self.fn = A.RandomCrop(*size)


@Transforms.register("a_shift_scale_rotate")
class AShiftScaleRotate(ATransforms):
    def __init__(
        self,
        p: float = 0.5,
        border_mode: Optional[int] = None,
        *,
        label_alias: Optional[str] = None,
    ):
        super().__init__(label_alias=label_alias)
        if border_mode is None:
            if cv2 is None:
                raise ValueError("`cv2` is needed for `AShiftScaleRotate`")
            border_mode = cv2.BORDER_REFLECT_101
        self.fn = A.ShiftScaleRotate(border_mode=border_mode, p=p)


@Transforms.register("a_hflip")
class AHFlip(ATransforms):
    def __init__(self, p: float = 0.5, *, label_alias: Optional[str] = None):
        super().__init__(label_alias=label_alias)
        self.fn = A.HorizontalFlip(p=p)


@Transforms.register("a_vflip")
class AVFlip(ATransforms):
    def __init__(self, p: float = 0.5, *, label_alias: Optional[str] = None):
        super().__init__(label_alias=label_alias)
        self.fn = A.VerticalFlip(p=p)


@Transforms.register("a_normalize")
class ANormalize(ATransforms):
    def __init__(
        self,
        mean: Tuple[float, float, float] = (0.485, 0.456, 0.406),
        std: Tuple[float, float, float] = (0.229, 0.224, 0.225),
        max_pixel_value: float = 1.0,
        p: float = 1.0,
        *,
        label_alias: Optional[str] = None,
    ):
        super().__init__(label_alias=label_alias)
        self.fn = A.Normalize(mean, std, max_pixel_value, p=p)


@Transforms.register("a_rgb_shift")
class ARGBShift(ATransforms):
    def __init__(
        self,
        r_shift_limit: float = 0.08,
        g_shift_limit: float = 0.08,
        b_shift_limit: float = 0.08,
        p: float = 0.5,
        *,
        label_alias: Optional[str] = None,
    ):
        super().__init__(label_alias=label_alias)
        self.fn = A.RGBShift(r_shift_limit, g_shift_limit, b_shift_limit, p=p)


@Transforms.register("a_solarize")
class ASolarize(ATransforms):
    def __init__(
        self,
        threshold: float = 0.5,
        p: float = 0.5,
        *,
        label_alias: Optional[str] = None,
    ):
        super().__init__(label_alias=label_alias)
        self.fn = A.Solarize(threshold, p=p)


@Transforms.register("a_gaussian_blur")
class AGaussianBlur(ATransforms):
    def __init__(
        self,
        blur_limit: Tuple[int, int] = (3, 7),
        sigma_limit: int = 0,
        p: float = 0.5,
        *,
        label_alias: Optional[str] = None,
    ):
        super().__init__(label_alias=label_alias)
        self.fn = A.GaussianBlur(blur_limit, sigma_limit, p=p)


@Transforms.register("a_hue_saturation")
class AHueSaturationValue(ATransforms):
    def __init__(
        self,
        hue_shift_limit: float = 0.08,
        sat_shift_limit: float = 0.12,
        val_shift_limit: float = 0.08,
        p: float = 0.5,
        *,
        label_alias: Optional[str] = None,
    ):
        super().__init__(label_alias=label_alias)
        self.fn = A.HueSaturationValue(
            hue_shift_limit,
            sat_shift_limit,
            val_shift_limit,
            p,
        )


@Transforms.register("a_brightness_contrast")
class ARandomBrightnessContrast(ATransforms):
    def __init__(
        self,
        brightness_limit: float = 0.2,
        contrast_limit: float = 0.2,
        brightness_by_max: bool = True,
        p: float = 0.5,
        *,
        label_alias: Optional[str] = None,
    ):
        super().__init__(label_alias=label_alias)
        self.fn = A.RandomBrightnessContrast(
            brightness_limit,
            contrast_limit,
            brightness_by_max,
            p,
        )


@Transforms.register("a_to_tensor")
class AToTensor(ATransforms):
    def __init__(
        self,
        transpose_mask: bool = True,
        *,
        label_alias: Optional[str] = None,
    ):
        super().__init__(label_alias=label_alias)
        self.fn = ToTensorV2(transpose_mask)


__all__ = [
    "AToRGB",
    "AToGray",
    "AResize",
    "ARandomCrop",
    "AShiftScaleRotate",
    "AHFlip",
    "AVFlip",
    "ANormalize",
    "ARGBShift",
    "ASolarize",
    "AGaussianBlur",
    "AHueSaturationValue",
    "ARandomBrightnessContrast",
    "AToTensor",
    "ATransforms",
]



import os
import json
import tempfile

import numpy as np

from abc import abstractmethod
from abc import ABCMeta
from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from typing import NamedTuple
from functools import partial
from cftool.misc import safe_execute
from cftool.misc import Saving
from cftool.misc import WithRegister
from cftool.array import squeeze
from cftool.array import is_float
from cftool.array import to_torch
from cftool.array import is_string
from cftool.types import np_dict_type
from cftool.types import tensor_dict_type

from .core import DLDataModule
from ..types import data_type
from ..types import sample_weights_type
from ..protocol import IDataset
from ..protocol import IDataLoader
from ..constants import INPUT_KEY
from ..constants import LABEL_KEY
from ..constants import PREDICTIONS_KEY
from ..constants import BATCH_INDICES_KEY
from ..misc.toolkit import ConfigMeta

try:
    from cfdata.tabular.api import TabularData
except:
    TabularData = None


ml_data_processors: Dict[str, Type["IMLDataProcessor"]] = {}


def get_weighted_indices(
    n: int,
    weights: Optional[np.ndarray],
    ensure_all_occur: bool = False,
) -> np.ndarray:
    indices = np.arange(n)
    if weights is not None:
        numbers = np.random.multinomial(n, weights)
        if ensure_all_occur:
            numbers += 1
        indices = indices.repeat(numbers)
    return indices


# protocols


class IMLPreProcessedData(NamedTuple):
    """
    * x_train (np.ndarray) : preprocessed training features.
    * y_train (Optional[np.ndarray]) : preprocessed training labels, could be None if not provided.
      * It is common that labels are not provided at inference time.
    * x_valid (Optional[np.ndarray]) : preprocessed validation features.
    * y_valid (Optional[np.ndarray]) : preprocessed validation labels, could be None if not provided.
    * input_dim (int) : input feature dim that the model will receive.
      * If not provided, `x_train.shape[-1]` will be used.
      * If `encoder` is provided, this setting will not represent the final input dim that your
      model will receive, because the `encoder` might 'expand' the dimension with some encoding methods.
    * num_history (Optional[int]) : number of history, useful in time series tasks.
      * If not provided, we will use the default value defined in the pipeline.
    * num_classes (Optional[int]) : number of classes, will be used as `output_dim` if `is_classification` is True & `output_dim` is not specified.
      * If not provided, we will use the default value defined in the pipeline.
    * is_classification (Optional[bool]) : whether current task is a classification task.
      * If not provided, we will use the default value defined in the pipeline.
    """

    x_train: np.ndarray
    y_train: Optional[np.ndarray] = None
    x_valid: Optional[np.ndarray] = None
    y_valid: Optional[np.ndarray] = None
    input_dim: Optional[int] = None
    num_history: Optional[int] = None
    num_classes: Optional[int] = None
    is_classification: Optional[bool] = None


class IMLBatch(NamedTuple):
    input: np.ndarray
    labels: Optional[np.ndarray]
    others: Optional[np_dict_type] = None


class IMLDataProcessor(WithRegister["IMLDataProcessor"], metaclass=ABCMeta):
    d = ml_data_processors

    is_ready: bool = False

    # abstract

    @abstractmethod
    def build_with(
        self,
        config: Dict[str, Any],
        x_train: Union[np.ndarray, str],
        y_train: Optional[Union[np.ndarray, str]],
        x_valid: Optional[Union[np.ndarray, str]],
        y_valid: Optional[Union[np.ndarray, str]],
    ) -> None:
        pass

    @abstractmethod
    def preprocess(
        self,
        config: Dict[str, Any],
        x_train: Union[np.ndarray, str],
        y_train: Optional[Union[np.ndarray, str]],
        x_valid: Optional[Union[np.ndarray, str]],
        y_valid: Optional[Union[np.ndarray, str]],
        *,
        for_inference: bool,
    ) -> IMLPreProcessedData:
        pass

    @abstractmethod
    def dumps(self) -> Any:
        pass

    @abstractmethod
    def loads(self, dumped: Any) -> None:
        pass

    # optional callbacks

    def get_num_samples(self, x: np.ndarray) -> Optional[int]:
        return None

    def fetch_batch(
        self,
        x: np.ndarray,
        y: Optional[np.ndarray],
        indices: Union[int, List[int], np.ndarray],
    ) -> IMLBatch:
        return IMLBatch(x[indices], None if y is None else y[indices])

    # changes can happen inplace
    def postprocess_batch(self, batch: np_dict_type) -> np_dict_type:
        return batch

    # changes can happen inplace
    def postprocess_results(
        self,
        forward: np_dict_type,
        *,
        return_classes: bool,
        binary_threshold: float,
        return_probabilities: bool,
    ) -> np_dict_type:
        return forward

    # api

    def to_pack(self) -> Dict[str, Any]:
        return dict(type=self.__identifier__, info=self.dumps())

    @classmethod
    def from_pack(cls, pack: Dict[str, Any]) -> "IMLDataProcessor":
        processor = cls.get(pack["type"])()
        safe_execute(processor.loads, dict(dumped=pack["info"]))
        processor.is_ready = True
        return processor


def register_ml_data_processor(name: str, *, allow_duplicate: bool = False) -> Callable:
    return IMLDataProcessor.register(name, allow_duplicate=allow_duplicate)


class IMLDataset(IDataset, metaclass=ABCMeta):
    @abstractmethod
    def __getitem__(self, item: Union[int, List[int], np.ndarray]) -> np_dict_type:
        pass


class IMLLoader(IDataLoader, metaclass=ABCMeta):
    callback: str

    cursor: int
    indices: np.ndarray

    data: IMLDataset
    shuffle: bool
    shuffle_backup: bool
    name: Optional[str]
    batch_size: int
    use_numpy: bool

    def __init__(
        self,
        data: IMLDataset,
        *,
        shuffle: bool,
        name: Optional[str] = None,
        batch_size: int = 128,
        sample_weights: Optional[np.ndarray] = None,
        use_numpy: bool = False,
    ):
        if sample_weights is not None and len(data) != len(sample_weights):
            raise ValueError(
                f"the number of data samples ({len(data)}) is not identical with "
                f"the number of sample weights ({len(sample_weights)})"
            )
        super().__init__(sample_weights=sample_weights)
        self.data = data
        self.shuffle = shuffle
        self.shuffle_backup = shuffle
        self.name = name
        self.batch_size = batch_size
        self.use_numpy = use_numpy

    def __iter__(self) -> "IMLLoader":
        self.cursor = 0
        self.indices = get_weighted_indices(len(self.data), self.sample_weights)
        if self.shuffle:
            np.random.shuffle(self.indices)
        return self

    def __next__(self) -> Union[np_dict_type, tensor_dict_type]:
        start = self.cursor
        if start >= len(self.data):
            raise StopIteration
        self.cursor += self.batch_size
        indices = self.indices[start : self.cursor]
        batch = self.data[indices]
        batch.setdefault(BATCH_INDICES_KEY, indices)
        if self.use_numpy:
            return batch
        return {k: None if v is None else to_torch(v) for k, v in batch.items()}

    def disable_shuffle(self) -> None:
        self.shuffle = False

    def recover_shuffle(self) -> None:
        self.shuffle = self.shuffle_backup

    def copy(self) -> "IMLLoader":
        return self.__class__(
            self.data,
            name=self.name,
            shuffle=self.shuffle,
            batch_size=self.batch_size,
            sample_weights=self.sample_weights,
        )


# internal


@IDataset.register("ml")
class MLDataset(IMLDataset):
    def __init__(
        self,
        x: np.ndarray,
        y: Optional[np.ndarray],
        processor: IMLDataProcessor,
        **others: np.ndarray,
    ):
        super().__init__()
        self.x = x
        self.y = y
        self.processor = processor
        self.others = others

    def __getitem__(self, item: Union[int, List[int], np.ndarray]) -> np_dict_type:
        ml_batch = self.processor.fetch_batch(self.x, self.y, item)
        batch = {
            INPUT_KEY: ml_batch.input,
            LABEL_KEY: ml_batch.labels,
        }
        if ml_batch.others is not None:
            for k, v in ml_batch.others.items():
                batch[k] = v
        for k, v in self.others.items():
            batch[k] = v[item]
        batch = self.processor.postprocess_batch(batch)
        return batch

    def __len__(self) -> int:
        return self.processor.get_num_samples(self.x) or len(self.x)


@IDataLoader.register("ml")
class MLLoader(IMLLoader):
    pass


class IMLPreProcessedXY(NamedTuple):
    x: np.ndarray
    y: Optional[np.ndarray]


class IMLData(DLDataModule, metaclass=ConfigMeta):
    config: Dict[str, Any]

    processor_type: str
    processor_info: Optional[Dict[str, Any]] = None

    train_data: MLDataset
    valid_data: Optional[MLDataset]

    data_files = [
        "x_train.npy",
        "y_train.npy",
        "x_valid.npy",
        "y_valid.npy",
    ]
    arguments_file = "arguments.json"

    input_dim: int
    num_history: int
    num_classes: Optional[int]
    is_classification: Optional[bool]

    processor: IMLDataProcessor
    train_others: Optional[np_dict_type]
    valid_others: Optional[np_dict_type]

    shuffle_train: bool
    shuffle_valid: bool
    batch_size: int
    valid_batch_size: int
    train_weights: Optional[np.ndarray]
    valid_weights: Optional[np.ndarray]

    use_numpy: bool
    for_inference: bool

    def __init__(
        self,
        x_train: Any,
        y_train: Optional[Any] = None,
        x_valid: Optional[Any] = None,
        y_valid: Optional[Any] = None,
        *,
        # processor
        processor: Optional[IMLDataProcessor] = None,
        # auxiliary data
        train_others: Optional[np_dict_type] = None,
        valid_others: Optional[np_dict_type] = None,
        # common
        num_history: int = 1,
        num_classes: Optional[int] = None,
        is_classification: Optional[bool] = None,
        # data loader
        shuffle_train: bool = True,
        shuffle_valid: bool = False,
        batch_size: int = 128,
        valid_batch_size: int = 512,
        # inference
        use_numpy: bool = False,
        for_inference: bool = False,
    ):
        pop_keys = [
            "x_train",
            "y_train",
            "x_valid",
            "y_valid",
            "processor",
            "train_others",
            "valid_others",
        ]
        for key in pop_keys:
            self.config.pop(key, None)
        self.x_train = x_train
        self.y_train = y_train
        self.x_valid = x_valid
        self.y_valid = y_valid
        self.train_others = train_others
        self.valid_others = valid_others
        self.num_history = num_history
        self.num_classes = num_classes
        self.is_classification = is_classification
        self.shuffle_train = shuffle_train
        self.shuffle_valid = shuffle_valid
        self.batch_size = batch_size
        self.valid_batch_size = valid_batch_size
        self.use_numpy = use_numpy
        self.for_inference = for_inference
        # processor
        if processor is None:
            processor = IMLDataProcessor.get(self.processor_type)()
        self.processor = processor
        if not processor.is_ready:
            self.build_processor()

    # processor

    @property
    def processor_build_config(self) -> Dict[str, Any]:
        return {}

    @property
    def processor_preprocess_config(self) -> Dict[str, Any]:
        return {}

    def build_processor(self) -> None:
        kw = dict(
            config=self.processor_build_config,
            x_train=self.x_train,
            y_train=self.y_train,
            x_valid=self.x_valid,
            y_valid=self.y_valid,
        )
        safe_execute(self.processor.build_with, kw)
        self.processor.is_ready = True

    def preprocess(
        self,
        x: Union[np.ndarray, str],
        y: Optional[Union[np.ndarray, str]],
    ) -> IMLPreProcessedXY:
        if not self.processor.is_ready:
            raise ValueError("`processor` should be ready before calling `preprocess`")
        kw = dict(
            config=self.processor_preprocess_config,
            x_train=x,
            y_train=y,
            x_valid=None,
            y_valid=None,
            for_inference=True,
        )
        res = safe_execute(self.processor.preprocess, kw)
        return IMLPreProcessedXY(res.x_train, res.y_train)

    # inheritance

    @property
    def info(self) -> Dict[str, Any]:
        if self.processor is None:
            raise ValueError(
                "`processor` should be provided before accessing `info`, "
                "did you forget to call the `prepare` method first?"
            )
        return {
            "type": self.__identifier__,
            "processor": self.processor,
            "input_dim": self.input_dim,
            "num_history": self.num_history,
            "num_classes": self.num_classes,
            "is_classification": self.is_classification,
        }

    def prepare(self, sample_weights: sample_weights_type) -> None:
        train_others = self.train_others or {}
        valid_others = self.valid_others or {}
        self.train_weights, self.valid_weights = _split_sw(sample_weights)
        preprocess_kw = dict(
            config=self.processor_preprocess_config,
            x_train=self.x_train,
            y_train=self.y_train,
            x_valid=self.x_valid,
            y_valid=self.y_valid,
            for_inference=self.for_inference,
        )
        final = safe_execute(self.processor.preprocess, preprocess_kw)
        data_info = dict(
            input_dim=final.input_dim or final.x_train.shape[-1],
            num_history=final.num_history,
            num_classes=final.num_classes,
            is_classification=final.is_classification,
        )
        for k, v in data_info.items():
            if v is not None:
                setattr(self, k, v)
        self.train_data = MLDataset(
            final.x_train,
            final.y_train,
            self.processor,
            **train_others,
        )
        if final.x_valid is None or final.y_valid is None:
            self.valid_data = None
        else:
            self.valid_data = MLDataset(
                final.x_valid,
                final.y_valid,
                self.processor,
                **valid_others,
            )

    def initialize(self) -> Tuple[MLLoader, Optional[MLLoader]]:
        if self.processor is None:
            raise ValueError(
                "`processor` should be provided before calling `initialize`, "
                "did you forget to call the `prepare` method first?"
            )
        train_loader = MLLoader(
            self.train_data,
            name=None if self.for_inference else "train",
            shuffle=self.shuffle_train,
            batch_size=self.batch_size,
            sample_weights=self.train_weights,
            use_numpy=self.use_numpy,
        )
        if self.valid_data is None:
            valid_loader = None
        else:
            # when `for_inference` is True, `valid_data` will always be `None`
            # so we don't need to condition `name` field here
            valid_loader = MLLoader(
                self.valid_data,
                name="valid",
                shuffle=self.shuffle_valid,
                batch_size=self.valid_batch_size,
                sample_weights=self.valid_weights,
                use_numpy=self.use_numpy,
            )
        return train_loader, valid_loader

    def _save_info(self, folder: str) -> None:
        info = self.info
        info["processor"] = self.processor.to_pack()
        Saving.save_dict(info, self.info_name, folder)

    def _save_data(self, data_folder: str) -> None:
        with open(os.path.join(data_folder, self.arguments_file), "w") as f:
            json.dump(self.config, f)
        all_data = [
            self.x_train,
            self.y_train,
            self.x_valid,
            self.y_valid,
        ]
        for data, file in zip(all_data, self.data_files):
            if data is not None:
                np.save(os.path.join(data_folder, file), data)
        if self.train_others is not None:
            for k, v in self.train_others.items():
                np.save(os.path.join(data_folder, f"{k}_train.npy"), v)
        if self.valid_others is not None:
            for k, v in self.valid_others.items():
                np.save(os.path.join(data_folder, f"{k}_valid.npy"), v)

    @classmethod
    def _get_load_arguments(cls, data_folder: str) -> Tuple[List[Any], Dict[str, Any]]:
        args = []
        for file in cls.data_files:
            path = os.path.join(data_folder, file)
            if not os.path.isfile(path):
                data = None
            else:
                data = np.load(path)
                if is_string(data):
                    data = data.item()
            args.append(data)
        with open(os.path.join(data_folder, cls.arguments_file), "r") as f:
            kwargs = json.load(f)
        train_others = {}
        valid_others = {}
        for file in os.listdir(data_folder):
            if file in cls.data_files:
                continue
            path = os.path.join(data_folder, file)
            if file.endswith("_train"):
                train_others[file.split("_train")[0]] = np.load(path)
            elif file.endswith("_valid"):
                valid_others[file.split("_valid")[0]] = np.load(path)
        if train_others:
            kwargs["train_others"] = train_others
        if valid_others:
            kwargs["valid_others"] = valid_others
        return args, kwargs

    @classmethod
    def _load_info(cls, folder: str) -> Dict[str, Any]:
        info = super()._load_info(folder)
        processor_pack = info.pop("processor")
        processor = IMLDataProcessor.from_pack(processor_pack)
        info["processor"] = processor
        return info

    @classmethod
    def _load(
        cls,
        data_folder: str,
        info: Dict[str, Any],
        sample_weights: sample_weights_type,
    ) -> "IMLData":
        args, kwargs = cls._get_load_arguments(data_folder)
        kwargs["processor"] = info["processor"]
        data = cls(*args, **kwargs)
        data.prepare(sample_weights)
        return data


def register_ml_data(name: str, *, allow_duplicate: bool = False) -> Callable:
    return IMLData.register(name, allow_duplicate=allow_duplicate)


# api


split_sw_type = Tuple[Optional[np.ndarray], Optional[np.ndarray]]


def _norm_sw(sample_weights: Optional[np.ndarray]) -> Optional[np.ndarray]:
    if sample_weights is None:
        return None
    return sample_weights / sample_weights.sum()


def _split_sw(sample_weights: sample_weights_type) -> split_sw_type:
    if sample_weights is None:
        train_weights = valid_weights = None
    else:
        if not isinstance(sample_weights, np.ndarray):
            train_weights, valid_weights = sample_weights
        else:
            train_weights, valid_weights = sample_weights, None
    train_weights, valid_weights = map(_norm_sw, [train_weights, valid_weights])
    return train_weights, valid_weights


@IMLDataProcessor.register("_internal.basic")
class _InternalBasicMLDataProcessor(IMLDataProcessor):
    def build_with(self) -> None:  # type: ignore
        pass

    def preprocess(  # type: ignore
        self,
        x_train: np.ndarray,
        y_train: Optional[np.ndarray],
        x_valid: Optional[np.ndarray],
        y_valid: Optional[np.ndarray],
    ) -> IMLPreProcessedData:
        return IMLPreProcessedData(x_train, y_train, x_valid, y_valid)

    def dumps(self) -> Any:
        return {}

    def loads(self) -> None:  # type: ignore
        pass


@DLDataModule.register("ml")
class MLData(IMLData):
    processor_type = "_internal.basic"


class MLInferenceData(MLData):
    def __init__(
        self,
        x: np.ndarray,
        y: Optional[np.ndarray] = None,
        *,
        shuffle: bool = False,
    ):
        super().__init__(
            x,
            y,
            shuffle_train=shuffle,
            for_inference=True,
        )
        self.prepare(None)


@IMLDataProcessor.register("_internal.carefree")
class _InternalCarefreeMLDataProcessor(IMLDataProcessor):
    cf_data: TabularData

    tmp_cf_data_name = ".tmp_cf_data"
    full_cf_data_name = "cf_data"

    def build_with(  # type: ignore
        self,
        config: Dict[str, Any],
        x_train: Union[np.ndarray, str],
        y_train: Optional[Union[np.ndarray, str]],
    ) -> None:
        self.cf_data = TabularData(**config["data_config"])
        self.cf_data.read(x_train, y_train, **config["read_config"])

    def preprocess(
        self,
        config: Dict[str, Any],
        x_train: Union[np.ndarray, str],
        y_train: Optional[Union[np.ndarray, str]],
        x_valid: Optional[Union[np.ndarray, str]],
        y_valid: Optional[Union[np.ndarray, str]],
        *,
        for_inference: bool,
    ) -> IMLPreProcessedData:
        if for_inference:
            x_train, y_train = self.cf_data.transform(
                x_train,
                y_train,
                contains_labels=config["contains_labels"],
            ).xy
            return IMLPreProcessedData(x_train, y_train)
        # split data
        if x_valid is not None:
            train_cf_data = self.cf_data
            valid_cf_data = self.cf_data.copy_to(x_valid, y_valid)
        else:
            if isinstance(config["valid_split"], int):
                split = config["valid_split"]
            else:
                num_data = len(self.cf_data)
                if isinstance(config["valid_split"], float):
                    split = int(round(config["valid_split"] * num_data))
                else:
                    default_split = 0.1
                    num_split = int(round(default_split * num_data))
                    num_split = max(config["min_valid_split"], num_split)
                    max_split = int(round(num_data * config["max_valid_split_ratio"]))
                    max_split = min(max_split, config["max_valid_split"])
                    split = min(num_split, max_split)
            if split <= 0:
                train_cf_data = self.cf_data
                valid_cf_data = None
            else:
                rs = self.cf_data.split(split, order=config["valid_split_order"])
                train_cf_data = rs.remained
                valid_cf_data = rs.split
        # process data
        if train_cf_data.processed is not None:
            x_train, y_train = train_cf_data.processed.xy
        else:
            x_train, y_train = train_cf_data.transform(
                x_train,
                y_train,
                contains_labels=config["contains_labels"],
            ).xy
        if valid_cf_data is None:
            x_valid = y_valid = None
        else:
            x_valid, y_valid = valid_cf_data.processed.xy
        is_classification = config["is_classification"]
        if is_classification is None:
            is_classification = train_cf_data.is_clf
        return IMLPreProcessedData(
            x_train,
            y_train,
            x_valid,
            y_valid,
            num_classes=train_cf_data.num_classes,
            is_classification=is_classification,
        )

    def dumps(self) -> Any:
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_name = os.path.join(tmp_dir, self.tmp_cf_data_name)
            self.cf_data.save(tmp_name, retain_data=False)
            zip_file = f"{tmp_name}.zip"
            with open(zip_file, "rb") as f:
                cf_data_bytes = f.read()
            os.remove(zip_file)
        return cf_data_bytes

    def loads(self, dumped: Any) -> None:
        if TabularData is None:
            msg = "`carefree-data` needs to be installed to load `MLCarefreeData`"
            raise ValueError(msg)
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_name = os.path.join(tmp_dir, self.tmp_cf_data_name)
            zip_file = f"{tmp_name}.zip"
            with open(zip_file, "wb") as f:
                f.write(dumped)
            self.cf_data = TabularData.load(tmp_name)
            os.remove(zip_file)

    # callbacks

    def postprocess_results(  # type: ignore
        self,
        forward: np_dict_type,
        *,
        return_classes: bool,
        return_probabilities: bool,
    ) -> np_dict_type:
        cf_data = self.cf_data
        is_clf = cf_data.is_clf
        if is_clf and return_probabilities:
            return forward
        fn = partial(cf_data.recover_labels, inplace=True)
        recovered = {}
        for k, v in forward.items():
            if is_clf and k != PREDICTIONS_KEY:
                continue
            if is_clf ^ is_float(v):
                if v.shape[1] == 1:
                    v = fn(v)
                else:
                    v = squeeze(np.apply_along_axis(fn, axis=0, arr=v))
            if is_clf and return_classes and is_float(v):
                v = v.astype(int)
            recovered[k] = v
        return recovered


@DLDataModule.register("ml.carefree")
class MLCarefreeData(IMLData, metaclass=ConfigMeta):
    processor_type = "_internal.carefree"

    processor: _InternalCarefreeMLDataProcessor

    def __init__(
        self,
        x_train: data_type,
        y_train: data_type = None,
        x_valid: data_type = None,
        y_valid: data_type = None,
        *,
        # processor
        processor: Optional[_InternalCarefreeMLDataProcessor] = None,
        data_config: Optional[Dict[str, Any]] = None,
        read_config: Optional[Dict[str, Any]] = None,
        # auxiliary data
        train_others: Optional[np_dict_type] = None,
        valid_others: Optional[np_dict_type] = None,
        # common
        num_history: int = 1,
        num_classes: Optional[int] = None,
        is_classification: Optional[bool] = None,
        # valid split
        valid_split: Optional[Union[int, float]] = None,
        min_valid_split: int = 100,
        max_valid_split: int = 10000,
        max_valid_split_ratio: float = 0.5,
        valid_split_order: str = "auto",
        # data loader
        shuffle_train: bool = True,
        shuffle_valid: bool = False,
        batch_size: int = 128,
        valid_batch_size: int = 512,
        # inference
        use_numpy: bool = False,
        for_inference: bool = False,
        contains_labels: bool = True,
    ):
        self.data_config = data_config or {}
        self.read_config = read_config or {}
        self.valid_split = valid_split
        self.min_valid_split = min_valid_split
        self.max_valid_split = max_valid_split
        self.max_valid_split_ratio = max_valid_split_ratio
        self.valid_split_order = valid_split_order
        self.contains_labels = contains_labels
        super().__init__(
            x_train,
            y_train,
            x_valid,
            y_valid,
            processor=processor,
            train_others=train_others,
            valid_others=valid_others,
            num_history=num_history,
            num_classes=num_classes,
            is_classification=is_classification,
            shuffle_train=shuffle_train,
            shuffle_valid=shuffle_valid,
            batch_size=batch_size,
            valid_batch_size=valid_batch_size,
            use_numpy=use_numpy,
            for_inference=for_inference,
        )

    @property
    def processor_build_config(self) -> Dict[str, Any]:
        return dict(
            data_config=self.data_config,
            read_config=self.read_config,
        )

    @property
    def processor_preprocess_config(self) -> Dict[str, Any]:
        return dict(
            contains_labels=self.contains_labels,
            valid_split=self.valid_split,
            min_valid_split=self.min_valid_split,
            max_valid_split=self.max_valid_split,
            max_valid_split_ratio=self.max_valid_split_ratio,
            valid_split_order=self.valid_split_order,
            is_classification=self.is_classification,
        )

    @classmethod
    def make_with(
        cls,
        *args: Any,
        is_classification: Optional[bool] = None,
        cf_data_config: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> "MLCarefreeData":
        if TabularData is None:
            msg = "`carefree-data` needs to be installed for making `MLCarefreeData`"
            raise ValueError(msg)
        if cf_data_config is None:
            cf_data_config = {}
        cf_data_config["default_categorical_process"] = "identical"
        if is_classification is not None:
            cf_data_config["task_type"] = "clf" if is_classification else "reg"
        kwargs["is_classification"] = is_classification
        kwargs["data_config"] = cf_data_config
        return cls(*args, **kwargs)


__all__ = [
    "get_weighted_indices",
    "register_ml_data",
    "register_ml_data_processor",
    "IMLBatch",
    "IMLDataProcessor",
    "IMLPreProcessedData",
    "IMLDataset",
    "IMLLoader",
    "IMLData",
    "MLDataset",
    "MLLoader",
    "MLData",
    "MLInferenceData",
    "MLCarefreeData",
]



import copy
import torch

import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Iterator
from typing import Optional
from functools import wraps
from torch.nn import Module
from torch.nn import ModuleList
from cftool.types import tensor_dict_type


def get_clones(
    module: Module,
    n: int,
    *,
    return_list: bool = False,
) -> Union[ModuleList, List[Module]]:
    module_list = [module]
    for _ in range(n - 1):
        module_list.append(copy.deepcopy(module))
    if return_list:
        return module_list
    return ModuleList(module_list)


def reuse_fn(f: Callable) -> Callable:
    cache = None

    @wraps(f)
    def cached_fn(*args: Any, _cache: bool = True, **kwargs: Any) -> Callable:
        if not _cache:
            return f(*args, **kwargs)
        nonlocal cache
        if cache is not None:
            return cache
        cache = f(*args, **kwargs)
        return cache

    return cached_fn


class Lambda(Module):
    def __init__(self, fn: Callable, name: str = None):
        super().__init__()
        self.name = name
        self.fn = fn

    def extra_repr(self) -> str:
        return "" if self.name is None else self.name

    def forward(self, *args: Any, **kwargs: Any) -> Any:
        return self.fn(*args, **kwargs)


class EMA(Module):
    def __init__(
        self,
        decay: float,
        named_parameters: List[Tuple[str, nn.Parameter]],
        *,
        use_num_updates: bool = False,
    ):
        super().__init__()
        self._cache: tensor_dict_type = {}
        self._decay = decay
        self._named_parameters = named_parameters
        for name, param in self.tgt_params:
            self.register_buffer(name, param.data.clone())
        num_updates = torch.tensor(0 if use_num_updates else -1, dtype=torch.int)
        self.register_buffer("num_updates", num_updates)

    @property
    def tgt_params(self) -> Iterator[Tuple[str, nn.Parameter]]:
        return map(
            lambda pair: (pair[0].replace(".", "_"), pair[1]),
            self._named_parameters,
        )

    def forward(self) -> None:
        if not self.training:
            raise ValueError("should not update `EMA` at inference stage")
        if self.num_updates < 0:
            decay = self._decay
        else:
            self.num_updates += 1
            decay = min(self._decay, (1 + self.num_updates) / (10 + self.num_updates))
        for name, param in self.tgt_params:
            ema_attr = getattr(self, name)
            ema = (1.0 - decay) * param.data + decay * ema_attr
            setattr(self, name, ema.clone())

    def train(self, mode: bool = True) -> "EMA":
        super().train(mode)
        if mode:
            for name, param in self.tgt_params:
                cached = self._cache.pop(name, None)
                if cached is not None:
                    param.data = cached
        else:
            for name, param in self.tgt_params:
                if name not in self._cache:
                    self._cache[name] = param.data
                param.data = getattr(self, name).clone()
        return self

    def extra_repr(self) -> str:
        max_str_len = max(len(name) for name, _ in self.tgt_params)
        return "\n".join(
            [f"(0): decay_rate={self._decay}\n(1): Params("]
            + [
                f"  {name:<{max_str_len}s} - Tensor({list(param.shape)})"
                for name, param in self.tgt_params
            ]
            + [")"]
        )


class MTL(Module):
    def __init__(
        self,
        num_tasks: int,
        method: Optional[str] = None,
    ):
        super().__init__()
        self._n_task, self._method = num_tasks, method
        if method is None or method == "naive":
            pass
        elif method == "softmax":
            self.w = torch.nn.Parameter(torch.ones(num_tasks))
        else:
            raise NotImplementedError(f"MTL method '{method}' not implemented")
        self.registered = False
        self._slice: Optional[int] = None
        self._registered: Dict[str, int] = {}

    def register(self, names: List[str]) -> None:
        if self.registered:
            raise ValueError("re-register is not permitted")
        self._rev_registered = {}
        for name in sorted(names):
            idx = len(self._registered)
            self._registered[name], self._rev_registered[idx] = idx, name
        self._slice, self.registered = len(names), True
        if self._slice > self._n_task:
            raise ValueError("registered names are more than n_task")

    def forward(
        self,
        loss_dict: tensor_dict_type,
        naive: bool = False,
    ) -> Tensor:
        if not self.registered:
            raise ValueError("losses need to be registered")
        if naive or self._method is None:
            return self._naive(loss_dict)
        return getattr(self, f"_{self._method}")(loss_dict)

    @staticmethod
    def _naive(loss_dict: tensor_dict_type) -> Tensor:
        return sum(loss_dict.values())  # type: ignore

    def _softmax(self, loss_dict: tensor_dict_type) -> Tensor:
        assert self._slice is not None
        w = self.w if self._slice == self._n_task else self.w[: self._slice]
        softmax_w = nn.functional.softmax(w, dim=0)
        losses = []
        for key, loss in loss_dict.items():
            idx = self._registered.get(key)
            losses.append(loss if idx is None else loss * softmax_w[idx])
        final_loss: Tensor = sum(losses)  # type: ignore
        return final_loss * self._slice

    def extra_repr(self) -> str:
        method = "naive" if self._method is None else self._method
        return f"n_task={self._n_task}, method='{method}'"


class ApplyTanhMixin:
    apply_tanh: bool

    def postprocess(self, net: Tensor, apply_tanh: Optional[bool]) -> Tensor:
        if apply_tanh is None:
            apply_tanh = self.apply_tanh
        if apply_tanh:
            net = torch.tanh(net)
        return net


__all__ = [
    "get_clones",
    "reuse_fn",
    "Lambda",
    "EMA",
    "MTL",
    "ApplyTanhMixin",
]



import torch

import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Optional

from ....misc.internal_ import register_module


class ResidualDenseBlock(nn.Module):
    def __init__(self, in_channels: int, grow_channels: int):
        super().__init__()
        self.conv1 = nn.Conv2d(in_channels, grow_channels, 3, 1, 1)
        self.conv2 = nn.Conv2d(in_channels + grow_channels, grow_channels, 3, 1, 1)
        self.conv3 = nn.Conv2d(in_channels + 2 * grow_channels, grow_channels, 3, 1, 1)
        self.conv4 = nn.Conv2d(in_channels + 3 * grow_channels, grow_channels, 3, 1, 1)
        self.conv5 = nn.Conv2d(in_channels + 4 * grow_channels, in_channels, 3, 1, 1)
        self.lrelu = nn.LeakyReLU(negative_slope=0.2, inplace=True)

    def forward(self, net: Tensor) -> Tensor:
        inp = net
        net1 = self.lrelu(self.conv1(inp))
        net2 = self.lrelu(self.conv2(torch.cat((inp, net1), 1)))
        net3 = self.lrelu(self.conv3(torch.cat((inp, net1, net2), 1)))
        net4 = self.lrelu(self.conv4(torch.cat((inp, net1, net2, net3), 1)))
        net = self.conv5(torch.cat((inp, net1, net2, net3, net4), 1))
        return net * 0.2 + inp


class RRDB(nn.Module):
    def __init__(self, in_channels: int, grow_channels: int):
        super().__init__()
        self.rdb1 = ResidualDenseBlock(in_channels, grow_channels)
        self.rdb2 = ResidualDenseBlock(in_channels, grow_channels)
        self.rdb3 = ResidualDenseBlock(in_channels, grow_channels)

    def forward(self, net: Tensor) -> Tensor:
        inp = net
        net = self.rdb1(net)
        net = self.rdb2(net)
        net = self.rdb3(net)
        return net * 0.2 + inp


def pixel_unshuffle(net: Tensor, scale: int) -> Tensor:
    b, c, h, w = net.shape
    out_channel = c * (scale**2)
    assert h % scale == 0 and w % scale == 0
    h = h // scale
    w = w // scale
    net = net.view(b, c, h, scale, w, scale)
    net = net.permute(0, 1, 3, 5, 2, 4).contiguous().view(b, out_channel, h, w)
    return net


@register_module("rrdb")
class RRDBNet(nn.Module):
    def __init__(
        self,
        in_channels: int = 3,
        out_channels: Optional[int] = None,
        *,
        scale: int = 4,
        latent_channels: int = 64,
        grow_channels: int = 32,
        num_layers: int = 23,
    ):
        super().__init__()
        self.scale = scale
        out_channels = out_channels or in_channels
        if scale == 2:
            in_channels = in_channels * 4
        elif scale == 1:
            in_channels = in_channels * 16
        self.conv_first = nn.Conv2d(in_channels, latent_channels, 3, 1, 1)
        make_layer = lambda: RRDB(latent_channels, grow_channels)
        self.body = nn.Sequential(*[make_layer() for _ in range(num_layers)])
        self.conv_body = nn.Conv2d(latent_channels, latent_channels, 3, 1, 1)
        # upsample
        self.conv_up1 = nn.Conv2d(latent_channels, latent_channels, 3, 1, 1)
        self.conv_up2 = nn.Conv2d(latent_channels, latent_channels, 3, 1, 1)
        self.conv_hr = nn.Conv2d(latent_channels, latent_channels, 3, 1, 1)
        self.conv_last = nn.Conv2d(latent_channels, out_channels, 3, 1, 1)
        # misc
        self.lrelu = nn.LeakyReLU(negative_slope=0.2, inplace=True)
        self.upsample = lambda net: F.interpolate(net, scale_factor=2, mode="nearest")

    def forward(self, net: Tensor) -> Tensor:
        if self.scale == 2:
            net = pixel_unshuffle(net, scale=2)
        elif self.scale == 1:
            net = pixel_unshuffle(net, scale=4)
        net = self.conv_first(net)
        body = self.conv_body(self.body(net))
        net = net + body
        # upsample
        net = self.lrelu(self.conv_up1(self.upsample(net)))
        net = self.lrelu(self.conv_up2(self.upsample(net)))
        net = self.conv_last(self.lrelu(self.conv_hr(net)))
        return net


__all__ = [
    "RRDBNet",
]



from ...pipeline import DLPipeline
from ...misc.internal_.inference import NLPInference


@DLPipeline.register("nlp")
class NLPPipeline(DLPipeline):
    inference: NLPInference
    inference_base = NLPInference


__all__ = [
    "NLPPipeline",
]



import os
import torch

import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Union
from typing import Optional
from cftool.misc import update_dict
from cftool.misc import check_requires
from cftool.misc import shallow_copy_dict
from cftool.types import tensor_dict_type
from torch.utils.data import Dataset

from .core import DLLoader
from .core import DLDataset
from .core import DataLoader
from .core import DLDataModule
from ..types import sample_weights_type
from ..constants import INPUT_KEY
from ..constants import LABEL_KEY
from ..constants import ORIGINAL_LABEL_KEY
from ..misc.toolkit import eval_context


class TensorDataset(Dataset):
    def __init__(
        self,
        x: Tensor,
        y: Optional[Tensor],
        others: Optional[tensor_dict_type] = None,
    ):
        self.x = x
        self.y = y
        self.others = others

    def __getitem__(self, index: int) -> tensor_dict_type:
        item = {INPUT_KEY: self.x[index]}
        if self.y is not None:
            label = self.y[index]
            item.update({LABEL_KEY: label, ORIGINAL_LABEL_KEY: label})
        if self.others is not None:
            for k, v in self.others.items():
                item[k] = v[index]
        return item

    def __len__(self) -> int:
        return len(self.x)


class TensorDictDataset(Dataset):
    def __init__(
        self,
        x: tensor_dict_type,
        y: Optional[Tensor],
    ):
        self.x = x
        self.y = y
        self.x_keys = sorted(self.x)

    def __getitem__(self, index: int) -> tensor_dict_type:
        item = {k: self.x[k][index] for k in self.x_keys}
        if self.y is not None:
            label = self.y[index]
            item.update({LABEL_KEY: label, ORIGINAL_LABEL_KEY: label})
        return item

    def __len__(self) -> int:
        return len(self.x[self.x_keys[0]])


# api


@DLDataModule.register("tensor")
class TensorData(DLDataModule):
    files = [
        "x_train.pt",
        "y_train.pt",
        "x_valid.pt",
        "y_valid.pt",
        "train_others.pt",
        "valid_others.pt",
    ]

    def __init__(
        self,
        x_train: Tensor,
        y_train: Optional[Tensor] = None,
        x_valid: Optional[Tensor] = None,
        y_valid: Optional[Tensor] = None,
        train_others: Optional[tensor_dict_type] = None,
        valid_others: Optional[tensor_dict_type] = None,
        *,
        shuffle: bool = True,
        batch_size: int = 64,
        num_workers: int = 0,
    ):
        self.x_train = x_train
        self.y_train = y_train
        self.x_valid = x_valid
        self.y_valid = y_valid
        self.train_others = train_others
        self.valid_others = valid_others
        self.kw = dict(batch_size=batch_size, shuffle=shuffle, num_workers=num_workers)

    @property
    def info(self) -> Dict[str, Any]:
        return self.kw

    # TODO : support sample weights
    def prepare(self, sample_weights: sample_weights_type) -> None:
        def _get_data(x: Any, y: Any, others: Any) -> DLDataset:
            return DLDataset(TensorDataset(x, y, others))

        self.train_data = _get_data(self.x_train, self.y_train, self.train_others)
        if self.x_valid is None:
            self.valid_data = None
        else:
            self.valid_data = _get_data(self.x_valid, self.y_valid, self.valid_others)

    def initialize(self) -> Tuple[DLLoader, Optional[DLLoader]]:
        train_loader = DLLoader(DataLoader(self.train_data, **self.kw))  # type: ignore
        if self.valid_data is None:
            valid_loader = None
        else:
            valid_loader = DLLoader(DataLoader(self.valid_data, **self.kw))  # type: ignore
        return train_loader, valid_loader

    def _save_data(self, data_folder: str) -> None:
        all_data = [self.x_train, self.y_train, self.x_valid, self.y_valid]
        all_data += [self.train_others, self.valid_others]
        for data, file in zip(all_data, self.files):
            if data is not None:
                torch.save(data, os.path.join(data_folder, file))

    @classmethod
    def _load(
        cls,
        data_folder: str,
        info: Dict[str, Any],
        sample_weights: sample_weights_type,
    ) -> "TensorData":
        args = []
        for file in cls.files:
            path = os.path.join(data_folder, file)
            args.append(None if not os.path.isfile(path) else torch.load(path))
        data = cls(*args, **info)
        data.prepare(sample_weights)
        return data


@DLDataModule.register("tensor_dict")
class TensorDictData(DLDataModule):
    files = ["x_train.pt", "y_train.pt", "x_valid.pt", "y_valid.pt"]

    def __init__(
        self,
        x_train: tensor_dict_type,
        y_train: Optional[Tensor] = None,
        x_valid: Optional[tensor_dict_type] = None,
        y_valid: Optional[Tensor] = None,
        *,
        shuffle: bool = True,
        batch_size: int = 64,
        num_workers: int = 0,
    ):
        self.x_train = x_train
        self.y_train = y_train
        self.x_valid = x_valid
        self.y_valid = y_valid
        self.kw = dict(batch_size=batch_size, shuffle=shuffle, num_workers=num_workers)

    @property
    def info(self) -> Dict[str, Any]:
        return self.kw

    # TODO : support sample weights
    def prepare(self, sample_weights: sample_weights_type) -> None:
        def _get_data(x: Any, y: Any) -> DLDataset:
            return DLDataset(TensorDictDataset(x, y))

        self.train_data = _get_data(self.x_train, self.y_train)
        if self.x_valid is None:
            self.valid_data = None
        else:
            self.valid_data = _get_data(self.x_valid, self.y_valid)

    def initialize(self) -> Tuple[DLLoader, Optional[DLLoader]]:
        train_loader = DLLoader(DataLoader(self.train_data, **self.kw))  # type: ignore
        if self.valid_data is None:
            valid_loader = None
        else:
            valid_loader = DLLoader(DataLoader(self.valid_data, **self.kw))  # type: ignore
        return train_loader, valid_loader

    def _save_data(self, data_folder: str) -> None:
        all_data = [self.x_train, self.y_train, self.x_valid, self.y_valid]
        for data, file in zip(all_data, self.files):
            if data is not None:
                torch.save(data, os.path.join(data_folder, file))

    @classmethod
    def _load(
        cls,
        data_folder: str,
        info: Dict[str, Any],
        sample_weights: sample_weights_type,
    ) -> "TensorData":
        args = []
        for file in cls.files:
            path = os.path.join(data_folder, file)
            args.append(None if not os.path.isfile(path) else torch.load(path))
        data = cls(*args, **info)  # type: ignore
        data.prepare(sample_weights)
        return data


class TensorInferenceData(TensorData):
    def __init__(
        self,
        x_train: Tensor,
        y_train: Optional[Tensor] = None,
        x_valid: Optional[Tensor] = None,
        y_valid: Optional[Tensor] = None,
        train_others: Optional[tensor_dict_type] = None,
        valid_others: Optional[tensor_dict_type] = None,
        *,
        shuffle: bool = False,
        batch_size: int = 64,
        num_workers: int = 0,
    ):
        super().__init__(
            x_train,
            y_train,
            x_valid,
            y_valid,
            train_others,
            valid_others,
            shuffle=shuffle,
            batch_size=batch_size,
            num_workers=num_workers,
        )
        self.prepare(None)


class TensorDictInferenceData(TensorDictData):
    def __init__(
        self,
        x_train: tensor_dict_type,
        y_train: Optional[Tensor] = None,
        x_valid: Optional[tensor_dict_type] = None,
        y_valid: Optional[Tensor] = None,
        *,
        shuffle: bool = False,
        batch_size: int = 64,
        num_workers: int = 0,
    ):
        super().__init__(
            x_train,
            y_train,
            x_valid,
            y_valid,
            shuffle=shuffle,
            batch_size=batch_size,
            num_workers=num_workers,
        )
        self.prepare(None)


@DLDataModule.register("dummy")
class DummyData(TensorData):
    def __init__(
        self,
        *,
        num_samples: int = 1,
        batch_size: int = 1,
        use_valid: bool = False,
    ):
        dummy = torch.zeros([num_samples, 1])
        x_valid = y_valid = dummy if use_valid else None
        super().__init__(dummy, dummy, x_valid, y_valid, batch_size=batch_size)


def predict_tensor_data(
    m: nn.Module,
    data: Union[TensorInferenceData, TensorDictInferenceData],
    *,
    batch_size: Optional[int] = None,
    **predict_kwargs: Any,
) -> Any:
    if batch_size is not None:
        data.kw["batch_size"] = batch_size
    loader = data.initialize()[0]
    results = []
    is_dict_result = False
    with eval_context(m):
        fn = m.forward
        for batch in loader:
            batch = shallow_copy_dict(batch)
            args: Tuple[Any, ...]
            if check_requires(fn, "batch"):
                args = ()
                kwargs = dict(batch=batch)
            else:
                kwargs = batch
                args = (kwargs.pop(INPUT_KEY),)
            update_dict(shallow_copy_dict(predict_kwargs), kwargs)
            rs = fn(*args, **kwargs)
            if not is_dict_result and isinstance(rs, dict):
                is_dict_result = True
            results.append(rs)
    if not is_dict_result:
        return torch.cat(results, dim=0)
    final = {}
    for k in results[0]:
        final[k] = torch.cat([rs[k] for rs in results], dim=0)
    return final


__all__ = [
    "TensorData",
    "TensorDictData",
    "TensorInferenceData",
    "TensorDictInferenceData",
    "DummyData",
    "predict_tensor_data",
]



import torch

import torch.nn as nn

from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Callable
from typing import Optional
from cftool.misc import filter_kw
from cftool.misc import safe_execute
from cftool.array import to_device
from cftool.types import tensor_dict_type

from .bases import ICustomLossModule
from ..protocol import _forward
from ..protocol import ITrainer
from ..protocol import StepOutputs
from ..protocol import TrainerState
from ..protocol import MetricsOutputs
from ..protocol import IDataLoader
from ..protocol import ModelWithCustomSteps
from ..constants import INPUT_KEY
from .protocols.ml import MERGED_KEY
from .protocols.ml import IMLCore


def register_ml_module(
    name: str,
    *,
    allow_duplicate: bool = False,
    pre_bases: Optional[List[type]] = None,
    post_bases: Optional[List[type]] = None,
) -> Callable:
    def _core(m: Type[nn.Module]) -> Type[nn.Module]:
        @IMLCore.register(name, allow_duplicate=allow_duplicate)
        class _(*bases):  # type: ignore
            def __init__(self, **kwargs: Any):
                super().__init__(**filter_kw(IMLCore.__init__, kwargs))
                kwargs["in_dim"] = kwargs["input_dim"]
                kwargs["out_dim"] = kwargs["output_dim"]
                self.core = safe_execute(m, kwargs)

            def _init_with_trainer(self, trainer: ITrainer) -> None:
                init_fn = getattr(self.core, "_init_with_trainer", None)
                if init_fn is not None:
                    init_fn(trainer)

            def forward(
                self,
                batch_idx: int,
                batch: tensor_dict_type,
                state: Optional[TrainerState] = None,
                **kwargs: Any,
            ) -> tensor_dict_type:
                key = MERGED_KEY if MERGED_KEY in batch else INPUT_KEY
                return _forward(self.core, batch_idx, batch, key, state, **kwargs)

            def train_step(  # type: ignore
                self,
                batch_idx: int,
                batch: tensor_dict_type,
                trainer: ITrainer,
                forward_kwargs: Dict[str, Any],
                loss_kwargs: Dict[str, Any],
            ) -> StepOutputs:
                train_step = getattr(self.core, "train_step", None)
                if train_step is not None:
                    return train_step(
                        batch_idx,
                        batch,
                        trainer,
                        forward_kwargs,
                        loss_kwargs,
                    )

            def evaluate_step(  # type: ignore
                self,
                loader: IDataLoader,
                portion: float,
                trainer: ITrainer,
            ) -> MetricsOutputs:
                evaluate_step = getattr(self.core, "evaluate_step", None)
                if evaluate_step is not None:
                    return evaluate_step(loader, portion, trainer)

        return m

    bases = (pre_bases or []) + [IMLCore] + (post_bases or [])
    return _core


def register_custom_loss_module(
    name: str,
    *,
    is_ml: bool,
    allow_duplicate: bool = False,
    pre_bases: Optional[List[type]] = None,
    post_bases: Optional[List[type]] = None,
) -> Callable:
    def _core(m: Type[ICustomLossModule]) -> Type[ICustomLossModule]:
        class _(*bases):  # type: ignore
            def __init__(self, *args: Any, **kwargs: Any):
                super().__init__()
                self.core = m(*args, **kwargs)

            def train_step(
                self,
                batch_idx: int,
                batch: tensor_dict_type,
                trainer: ITrainer,
                forward_kwargs: Dict[str, Any],
                loss_kwargs: Dict[str, Any],
            ) -> StepOutputs:
                with torch.cuda.amp.autocast(enabled=trainer.use_amp):
                    kwargs = dict(
                        batch_idx=batch_idx,
                        batch=batch,
                        trainer=trainer,
                        forward_kwargs=forward_kwargs,
                        loss_kwargs=loss_kwargs,
                    )
                    forward, loss_dict = safe_execute(self.core.get_losses, kwargs)
                trainer.post_loss_step(loss_dict)
                return StepOutputs(forward, {k: v.item() for k, v in loss_dict.items()})

            def evaluate_step(
                self,
                loader: IDataLoader,
                portion: float,
                trainer: ITrainer,
            ) -> MetricsOutputs:
                if trainer.metrics is not None:
                    outputs = trainer.inference.get_outputs(
                        loader,
                        portion=portion,
                        state=trainer.state,
                        metrics=trainer.metrics,
                        return_outputs=False,
                    )
                    assert outputs.metric_outputs is not None
                    return outputs.metric_outputs
                loss_items: Dict[str, List[float]] = {}
                for i, batch in enumerate(loader):
                    if i / len(loader) >= portion:
                        break
                    batch = to_device(batch, self.device)
                    kwargs = dict(
                        batch_idx=0,
                        batch=batch,
                        trainer=trainer,
                        forward_kwargs={},
                        loss_kwargs={},
                    )
                    _, losses = safe_execute(self.core.get_losses, kwargs)
                    for k, v in losses.items():
                        loss_items.setdefault(k, []).append(v.item())
                # gather
                mean_loss_items = {k: sum(v) / len(v) for k, v in loss_items.items()}
                score = trainer.weighted_loss_score(mean_loss_items)
                return MetricsOutputs(score, mean_loss_items)

            def forward(
                self,
                batch_idx: int,
                batch: tensor_dict_type,
                state: Optional[TrainerState] = None,
                **kwargs: Any,
            ) -> tensor_dict_type:
                key = MERGED_KEY if is_ml and MERGED_KEY in batch else INPUT_KEY
                return _forward(self.core, batch_idx, batch, key, state, **kwargs)

        if is_ml:
            register_ml_module(name, allow_duplicate=allow_duplicate)(_)
        else:
            ModelWithCustomSteps.register(name, allow_duplicate=allow_duplicate)(_)
        return m

    bases = (pre_bases or []) + [ModelWithCustomSteps] + (post_bases or [])
    return _core


__all__ = [
    "register_ml_module",
    "register_custom_loss_module",
]



import torch

import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import Dict
from typing import Optional
from cftool.types import tensor_dict_type

from ..general import EncoderDecoder
from ....constants import INPUT_KEY
from ....constants import LABEL_KEY
from ....constants import PREDICTIONS_KEY
from ...protocols.cv import GaussianGeneratorMixin
from ....losses.vae import MU_KEY
from ....losses.vae import LOG_VAR_KEY
from ....modules.blocks import Lambda
from ....modules.blocks import Activation
from ....modules.blocks import ChannelPadding
from ....misc.internal_.register import register_module


def reparameterize(mu: Tensor, log_var: Tensor) -> Tensor:
    std = torch.exp(0.5 * log_var)
    eps = torch.randn_like(std)
    return eps * std + mu


class VanillaVAEBase(nn.Module, GaussianGeneratorMixin):
    from_latent: nn.Module

    def __init__(
        self,
        is_id: bool,
        in_channels: int,
        out_channels: Optional[int] = None,
        target_downsample: Optional[int] = None,
        latent_padding_channels: Optional[int] = 16,
        num_classes: Optional[int] = None,
        *,
        latent: int = 128,
        img_size: Optional[int] = None,
        min_size: int = 2,
        num_downsample: Optional[int] = None,
        num_upsample: Optional[int] = None,
        latent_resolution: Optional[int] = None,
        encoder: str = "vanilla",
        decoder: str = "vanilla",
        encoder_config: Optional[Dict[str, Any]] = None,
        decoder_config: Optional[Dict[str, Any]] = None,
        output_activation: Optional[str] = "tanh",
    ):
        super().__init__()
        # encoder defaults
        if encoder_config is None:
            encoder_config = {}
        if encoder == "vanilla":
            encoder_config.setdefault("norm_type", "instance")
            encoder_config.setdefault("padding", "reflection")
        # decoder defaults
        if decoder_config is None:
            decoder_config = {}
        if decoder == "vanilla":
            decoder_config.setdefault("last_kernel_size", 7)
            decoder_config.setdefault("reduce_channel_on_upsample", True)
        # build
        self.generator = EncoderDecoder(
            is_id,
            in_channels,
            out_channels,
            target_downsample,
            latent_padding_channels,
            num_classes,
            latent=latent * 2,
            decoder_latent=latent,
            img_size=img_size,
            min_size=min_size,
            num_downsample=num_downsample,
            num_upsample=num_upsample,
            latent_resolution=latent_resolution,
            encoder=encoder,
            decoder=decoder,
            encoder_config=encoder_config,
            decoder_config=decoder_config,
        )
        self.latent_dim = latent
        self.num_classes = num_classes
        if output_activation is None:
            self.out = None
        else:
            self.out = Activation.make(output_activation)

    @property
    def can_reconstruct(self) -> bool:
        return True

    def decode(self, z: Tensor, *, labels: Optional[Tensor], **kwargs: Any) -> Tensor:
        if labels is None and self.num_classes is not None:
            labels = torch.randint(self.num_classes, [len(z), 1], device=z.device)
        batch = {INPUT_KEY: self.from_latent(z), LABEL_KEY: labels}
        net = self.generator.decode(batch, **kwargs)
        if self.out is not None:
            net = self.out(net)
        return net

    def forward(self, batch: tensor_dict_type, **kwargs: Any) -> tensor_dict_type:
        net = self.generator.encode(batch, **kwargs)
        mu, log_var = net.chunk(2, dim=1)
        net = reparameterize(mu, log_var)
        labels = None if self.num_classes is None else batch[LABEL_KEY]
        net = self.decode(net, labels=labels, **kwargs)
        return {PREDICTIONS_KEY: net, MU_KEY: mu, LOG_VAR_KEY: log_var}


@register_module("vae")
@register_module("vae1d")
class VanillaVAE1D(VanillaVAEBase):
    def __init__(
        self,
        in_channels: int,
        out_channels: Optional[int] = None,
        target_downsample: Optional[int] = None,
        latent_padding_channels: Optional[int] = 16,
        num_classes: Optional[int] = None,
        *,
        latent: int = 128,
        img_size: Optional[int] = None,
        min_size: int = 2,
        num_downsample: Optional[int] = None,
        num_upsample: Optional[int] = None,
        latent_resolution: Optional[int] = None,
        encoder: str = "vanilla",
        decoder: str = "vanilla",
        encoder_config: Optional[Dict[str, Any]] = None,
        decoder_config: Optional[Dict[str, Any]] = None,
        output_activation: Optional[str] = "tanh",
    ):
        super().__init__(
            True,
            in_channels,
            out_channels,
            target_downsample,
            latent_padding_channels,
            num_classes,
            latent=latent,
            img_size=img_size,
            min_size=min_size,
            num_downsample=num_downsample,
            num_upsample=num_upsample,
            latent_resolution=latent_resolution,
            encoder=encoder,
            decoder=decoder,
            encoder_config=encoder_config,
            decoder_config=decoder_config,
            output_activation=output_activation,
        )
        self.from_latent = nn.Identity()


@register_module("vae2d")
class VanillaVAE2D(VanillaVAEBase):
    def __init__(
        self,
        in_channels: int,
        out_channels: Optional[int] = None,
        target_downsample: Optional[int] = None,
        latent_padding_channels: Optional[int] = 16,
        num_classes: Optional[int] = None,
        *,
        latent: int = 128,
        img_size: Optional[int] = None,
        min_size: int = 2,
        num_downsample: Optional[int] = None,
        num_upsample: Optional[int] = None,
        latent_resolution: Optional[int] = None,
        encoder: str = "vanilla",
        decoder: str = "vanilla",
        encoder_config: Optional[Dict[str, Any]] = None,
        decoder_config: Optional[Dict[str, Any]] = None,
        output_activation: Optional[str] = "tanh",
    ):
        super().__init__(
            False,
            in_channels,
            out_channels,
            target_downsample,
            latent_padding_channels,
            num_classes,
            latent=latent,
            img_size=img_size,
            min_size=min_size,
            num_downsample=num_downsample,
            num_upsample=num_upsample,
            latent_resolution=latent_resolution,
            encoder=encoder,
            decoder=decoder,
            encoder_config=encoder_config,
            decoder_config=decoder_config,
            output_activation=output_activation,
        )
        self.latent_resolution = self.generator.latent_resolution
        assert self.latent_resolution is not None
        self.latent_dim = latent * self.latent_resolution**2
        shape = -1, latent, self.latent_resolution, self.latent_resolution
        blocks = [Lambda(lambda net: net.view(shape), f"reshape -> {shape}")]
        if latent_padding_channels is None:
            self.from_latent = blocks[0]
        else:
            blocks.append(
                ChannelPadding(
                    latent,
                    latent_padding_channels,
                    self.latent_resolution,
                )
            )
            self.from_latent = nn.Sequential(*blocks)


__all__ = [
    "reparameterize",
    "VanillaVAE1D",
    "VanillaVAE2D",
]



import torch

import numpy as np

from PIL import Image
from typing import List
from typing import Callable
from typing import Optional
from cftool.array import to_torch

from ...protocol import IImageExtractor
from ...protocol import ImageFolderLatentResponse
from ...cv.utils import predict_paths
from ...cv.utils import predict_folder
from ...cv.pipeline import CVPipeline
from ....data import TensorData
from ....types import texts_type
from ....protocol import IInference
from ....constants import INPUT_KEY
from ....constants import LATENT_KEY
from ....misc.toolkit import eval_context
from ....models.nlp.tokenizers import ITokenizer
from ....models.multimodal.clip import CLIP


class CLIPExtractor(IImageExtractor):
    clip: CLIP

    def __init__(self, m: CVPipeline, *, tokenizer: Optional[str] = None):
        self.m = m
        clip = m.model
        self.clip = clip
        self.img_size = clip.img_size
        self.transform = clip.get_transform()
        if tokenizer is None:
            if self.clip.context_length == 512:
                tokenizer = "clip.chinese"
            else:
                tokenizer = "clip"
        self.tokenizer = ITokenizer.make(tokenizer, {})

    @property
    def text_forward_fn(self) -> Callable:
        return lambda batch: {LATENT_KEY: self.clip.encode_text(batch[INPUT_KEY])}

    @property
    def image_forward_fn(self) -> Callable:
        return lambda batch: {LATENT_KEY: self.clip.encode_image(batch[INPUT_KEY])}

    def get_texts_latent(
        self,
        texts: texts_type,
        *,
        batch_size: int = 64,
        use_tqdm: bool = True,
    ) -> np.ndarray:
        if isinstance(texts, str):
            texts = [texts]
        texts_tensor = to_torch(self.tokenizer.tokenize(texts))
        loader = TensorData(texts_tensor, batch_size=batch_size).get_loaders()[0]
        original_forward = self.clip.forward
        self.clip.forward = lambda _, batch, *s, **kws: self.text_forward_fn(batch)  # type: ignore
        inference = IInference(model=self.clip)
        outputs = inference.get_outputs(loader, use_tqdm=use_tqdm)
        self.clip.forward = original_forward  # type: ignore
        return outputs.forward_results[LATENT_KEY]

    def get_image_latent(self, image: Image.Image) -> np.ndarray:
        inp = self.transform(image)[None].to(self.clip.device)
        with eval_context(self.clip):
            return self.clip.encode_image(inp).cpu().numpy()

    def get_paths_latent(
        self,
        image_paths: List[str],
        *,
        batch_size: int = 64,
        num_workers: int = 0,
        use_tqdm: bool = True,
    ) -> np.ndarray:
        original_forward = self.clip.forward
        self.clip.forward = lambda _, batch, *s, **kws: self.image_forward_fn(batch)  # type: ignore
        results = predict_paths(
            self.m,
            image_paths,
            batch_size=batch_size,
            num_workers=num_workers,
            transform=self.transform,
            use_tqdm=use_tqdm,
        )
        self.clip.forward = original_forward  # type: ignore
        return results.outputs[LATENT_KEY]

    def get_folder_latent(
        self,
        src_folder: str,
        *,
        batch_size: int,
        num_workers: int = 0,
        use_tqdm: bool = True,
    ) -> ImageFolderLatentResponse:
        original_forward = self.clip.forward
        self.clip.forward = lambda _, batch, *s, **kws: self.image_forward_fn(batch)  # type: ignore
        results = predict_folder(
            self.m,
            src_folder,
            batch_size=batch_size,
            num_workers=num_workers,
            transform=self.transform,
            use_tqdm=use_tqdm,
        )
        self.clip.forward = original_forward  # type: ignore
        return ImageFolderLatentResponse(results.outputs[LATENT_KEY], results.img_paths)

    def to_text_onnx(
        self,
        export_folder: str,
        *,
        num_samples: Optional[int] = None,
        onnx_file: str = "text.onnx",
    ) -> None:
        inp = to_torch(self.tokenizer.tokenize("Test."))
        self.clip.to_onnx(
            export_folder,
            {INPUT_KEY: inp},
            onnx_file=onnx_file,
            forward_fn=self.text_forward_fn,
            output_names=[LATENT_KEY],
            num_samples=num_samples,
        )

    def to_image_onnx(
        self,
        export_folder: str,
        *,
        num_samples: Optional[int] = None,
        onnx_file: str = "image.onnx",
    ) -> None:
        inp = torch.randn(1, 3, self.img_size, self.img_size)
        self.clip.to_onnx(
            export_folder,
            {INPUT_KEY: inp},
            onnx_file=onnx_file,
            forward_fn=self.image_forward_fn,
            output_names=[LATENT_KEY],
            num_samples=num_samples,
        )


__all__ = [
    "CLIPExtractor",
]



from torch import Tensor
from typing import Any
from typing import List
from typing import Optional

from .ddim import DDIMMixin
from .ddim import IGetEPSPred
from .ddim import IGetDenoised
from .protocol import ISampler


@ISampler.register("plms")
class PLMSSampler(DDIMMixin):
    def sample_step_core(
        self,
        image: Tensor,
        cond: Optional[Tensor],
        step: int,
        total_step: int,
        get_eps_pred: IGetEPSPred,
        get_denoised: IGetDenoised,
        *,
        eta: float,
        discretize: str,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
        temperature: float,
        noise_dropout: float,
        quantize_denoised: bool,
        **kwargs: Any,
    ) -> Tensor:
        eps_pred = get_eps_pred(image, self._ts)

        if len(self.old_eps) == 0:
            denoised = get_denoised(eps_pred)
            eps_pred_next = get_eps_pred(denoised, self._ts_next)
            eps_prime = 0.5 * (eps_pred + eps_pred_next)
        elif len(self.old_eps) == 1:
            eps_prime = 0.5 * (3.0 * eps_pred - self.old_eps[-1])
        elif len(self.old_eps) == 2:
            eps_prime = (
                23.0 * eps_pred - 16.0 * self.old_eps[-1] + 5.0 * self.old_eps[-2]
            ) / 12.0
        elif len(self.old_eps) >= 3:
            eps_prime = (
                55.0 * eps_pred
                - 59.0 * self.old_eps[-1]
                + 37.0 * self.old_eps[-2]
                - 9.0 * self.old_eps[-3]
            ) / 24.0
        else:
            raise ValueError("length of `old_eps` should be in {0, 1, 2, 3}")

        self.old_eps.append(eps_pred)
        if len(self.old_eps) >= 4:
            self.old_eps.pop(0)

        denoised = get_denoised(eps_prime)
        return denoised

    def _reset_buffers(
        self,
        eta: float,
        discretize: str,
        total_step: int,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
    ) -> None:
        super()._reset_buffers(
            eta,
            discretize,
            total_step,
            unconditional_cond,
            unconditional_guidance_scale,
        )
        self.old_eps: List[Tensor] = []


__all__ = [
    "PLMSSampler",
]



import os
import torch

import torch.nn.functional as F

from abc import abstractmethod
from abc import ABCMeta
from typing import Any
from typing import Dict
from typing import Optional
from typing import NamedTuple
from cftool.misc import print_warning
from cftool.types import tensor_dict_type

from ..protocol import ITrainer
from ..protocol import ILoss
from ..protocol import IDLModel
from ..constants import LOSS_KEY
from ..constants import LABEL_KEY
from ..constants import LATENT_KEY
from ..constants import PREDICTIONS_KEY
from ..misc.toolkit import set_requires_grad


class ICustomLossOutput(NamedTuple):
    forward_results: tensor_dict_type
    loss_dict: tensor_dict_type


class ICustomLossModule(torch.nn.Module):
    @abstractmethod
    def get_losses(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        trainer: ITrainer,
        forward_kwargs: Dict[str, Any],
        loss_kwargs: Dict[str, Any],
    ) -> ICustomLossOutput:
        pass

    @abstractmethod
    def forward(self, *args: Any, **kwargs: Any) -> Any:
        pass


class IBAKE(ICustomLossModule, metaclass=ABCMeta):
    lb: float
    bake_loss: ILoss
    w_ensemble: float
    is_classification: bool

    def _init_bake(
        self,
        lb: float,
        bake_loss: str,
        bake_loss_config: Optional[Dict[str, Any]],
        w_ensemble: float,
        is_classification: bool,
    ) -> None:
        self.lb = lb
        self.w_ensemble = w_ensemble
        self.is_classification = is_classification
        if bake_loss == "auto":
            bake_loss = "focal" if is_classification else "mae"
        self.bake_loss = ILoss.make(bake_loss, bake_loss_config or {})

    def get_losses(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        trainer: ITrainer,
        forward_kwargs: Dict[str, Any],
        loss_kwargs: Dict[str, Any],
    ) -> ICustomLossOutput:
        state = trainer.state
        forward_results = self(batch_idx, batch, state, **forward_kwargs)
        loss_dict = trainer.loss(forward_results, batch, state, **loss_kwargs)
        loss = loss_dict[LOSS_KEY]
        predictions = forward_results[PREDICTIONS_KEY]
        # BAKE
        latent = F.normalize(forward_results[LATENT_KEY])
        batch_size = latent.shape[0]
        eye = torch.eye(batch_size, device=latent.device)
        similarities = F.softmax(latent.mm(latent.t()) - eye * 1.0e9, dim=1)
        inv = (eye - self.w_ensemble * similarities).inverse()
        weights = (1.0 - self.w_ensemble) * inv
        if not self.is_classification:
            bake_inp = predictions
            soft_labels = weights.mm(predictions).detach()
        else:
            bake_inp = F.log_softmax(predictions)
            soft_labels = weights.mm(F.softmax(predictions, dim=1)).detach()
        bake_loss = self.bake_loss(
            {PREDICTIONS_KEY: bake_inp},
            {LABEL_KEY: soft_labels},
            state,
            **loss_kwargs,
        )[LOSS_KEY]
        loss_dict["bake"] = bake_loss
        loss_dict[LOSS_KEY] = loss + self.lb * bake_loss
        return ICustomLossOutput(forward_results, loss_dict)


class IRDropout(ICustomLossModule, metaclass=ABCMeta):
    lb: float
    is_classification: bool

    def get_losses(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        trainer: ITrainer,
        forward_kwargs: Dict[str, Any],
        loss_kwargs: Dict[str, Any],
    ) -> ICustomLossOutput:
        state = trainer.state
        fr1 = self(batch_idx, batch, state, **forward_kwargs)
        fr2 = self(batch_idx, batch, state, **forward_kwargs)
        loss_dict = trainer.loss(fr1, batch, state, **loss_kwargs)
        for k, v in trainer.loss(fr2, batch, state, **loss_kwargs).items():
            loss_dict[k] = loss_dict[k] + v
        loss = loss_dict[LOSS_KEY]
        # R-Dropout loss
        p1 = fr1[PREDICTIONS_KEY]
        p2 = fr2[PREDICTIONS_KEY]
        if not self.is_classification:
            r_dropout = ((p1 - p2) ** 2).mean()
        else:
            p1_loss = F.kl_div(F.log_softmax(p1, dim=-1), F.softmax(p2, dim=-1))
            p2_loss = F.kl_div(F.log_softmax(p2, dim=-1), F.softmax(p1, dim=-1))
            r_dropout = 0.5 * (p1_loss + p2_loss)
        loss_dict["r_dropout"] = r_dropout
        loss_dict[LOSS_KEY] = loss + self.lb * r_dropout
        return ICustomLossOutput(fr1, loss_dict)


class CascadeMixin:
    lv1_net: IDLModel
    lv2_net: IDLModel

    def _construct(
        self,
        lv1_model_name: str,
        lv2_model_name: str,
        lv1_model_config: Dict[str, Any],
        lv2_model_config: Dict[str, Any],
        lv1_model_ckpt_path: Optional[str],
        lv1_model_trainable: bool,
    ) -> None:
        self.lv1_net = IDLModel.make(lv1_model_name, lv1_model_config)
        if lv1_model_ckpt_path is not None:
            if not os.path.isfile(lv1_model_ckpt_path):
                print_warning(f"'{lv1_model_ckpt_path}' does not exist")
            else:
                state_dict = torch.load(lv1_model_ckpt_path, map_location="cpu")
                self.lv1_net.load_state_dict(state_dict)
        if not lv1_model_trainable:
            set_requires_grad(self.lv1_net, False)
        self.lv2_net = IDLModel.make(lv2_model_name, lv2_model_config)


__all__ = [
    "ICustomLossModule",
    "IBAKE",
    "IRDropout",
    "CascadeMixin",
]



import torch

from torch import Tensor


def append_dims(net: Tensor, ndim: int) -> Tensor:
    diff = ndim - net.ndim
    return net[(...,) + (None,) * diff]


def append_zero(net: Tensor) -> Tensor:
    return torch.cat([net, net.new_zeros([1])])



import torch

import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from cftool.types import tensor_dict_type

from .protocol import OneStageGANMixin
from .discriminators import DiscriminatorBase
from ..generator import UnetGenerator
from ....constants import LOSS_KEY
from ....constants import INPUT_KEY
from ....constants import LABEL_KEY
from ....constants import PREDICTIONS_KEY
from ....losses.gan import GANTarget
from ....misc.internal_.register import register_custom_module
from ....misc.internal_.register import CustomModule


@register_custom_module("pix2pix")
class Pix2Pix(OneStageGANMixin, CustomModule):  # type: ignore
    def __init__(
        self,
        in_channels: int = 3,
        out_channels: Optional[int] = None,
        num_downsample: int = 8,
        *,
        start_channels: int = 64,
        norm_type: Optional[str] = "batch",
        norm_kwargs: Optional[Dict[str, Any]] = None,
        use_dropout: bool = True,
        discriminator: str = "basic",
        discriminator_config: Optional[Dict[str, Any]] = None,
        gan_mode: str = "vanilla",
        gan_loss_config: Optional[Dict[str, Any]] = None,
        lb_l1: float = 100.0,
    ):
        super().__init__()
        self._initialize(gan_mode=gan_mode, gan_loss_config=gan_loss_config)
        self.generator = UnetGenerator(
            in_channels,
            out_channels,
            num_downsample,
            start_channels=start_channels,
            norm_type=norm_type,
            norm_kwargs=norm_kwargs,
            use_dropout=use_dropout,
        )
        if discriminator_config is None:
            discriminator_config = {}
        discriminator_config["in_channels"] = in_channels + self.generator.out_channels
        self.discriminator = DiscriminatorBase.make(
            discriminator,
            config=discriminator_config,
        )
        # l1 loss
        self.l1_loss = nn.L1Loss()
        self.lb_l1 = lb_l1

    @property
    def g_parameters(self) -> List[nn.Parameter]:
        return list(self.generator.parameters())

    @property
    def d_parameters(self) -> List[nn.Parameter]:
        return list(self.discriminator.parameters())

    def forward(self, net: Tensor) -> Tensor:
        return self.generator(net)

    def _g_losses(
        self,
        batch: tensor_dict_type,
        forward: tensor_dict_type,
    ) -> tensor_dict_type:
        src_domain = batch[INPUT_KEY]
        tgt_domain = batch[LABEL_KEY]
        fake = forward[PREDICTIONS_KEY]
        fake_concat = torch.cat([src_domain, fake], 1)
        pred_fake = self.discriminator(fake_concat)
        loss_g_gan = self.gan_loss(pred_fake, GANTarget(True))
        loss_g_l1 = self.l1_loss(fake, tgt_domain) * self.lb_l1
        loss_g = loss_g_gan + loss_g_l1
        return {LOSS_KEY: loss_g, "g_gan": loss_g_gan, "g_l1": loss_g_l1}

    def _d_losses(
        self,
        batch: tensor_dict_type,
        detached_forward: tensor_dict_type,
    ) -> tensor_dict_type:
        src_domain = batch[INPUT_KEY]
        tgt_domain = batch[LABEL_KEY]
        fake = detached_forward[PREDICTIONS_KEY]
        fake_concat = torch.cat([src_domain, fake], 1)
        pred_fake = self.discriminator(fake_concat)
        loss_d_fake = self.gan_loss(pred_fake, GANTarget(False))
        real_concat = torch.cat([src_domain, tgt_domain], 1)
        pred_real = self.discriminator(real_concat)
        loss_d_real = self.gan_loss(pred_real, GANTarget(True))
        d_loss = 0.5 * (loss_d_fake + loss_d_real)
        return {LOSS_KEY: d_loss, "d_real": loss_d_real, "d_fake": loss_d_fake}


__all__ = ["Pix2Pix"]



import torch

from torch import nn
from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from cftool.types import tensor_dict_type

from ....constants import LATENT_KEY
from ....constants import PREDICTIONS_KEY
from ..encoder.backbone import BackboneEncoder
from ...protocols.cv import ImageTranslatorMixin
from ....misc.toolkit import interpolate
from ....modules.blocks import get_conv_blocks
from ....modules.blocks import Conv2d
from ....misc.internal_.register import register_module


class UNetDecoderBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        skip_channels: int,
        out_channels: int,
        *,
        norm_type: Optional[str] = "batch",
    ):
        super().__init__()
        self.net = nn.Sequential(
            *get_conv_blocks(
                in_channels + skip_channels,
                out_channels,
                3,
                1,
                norm_type=norm_type,
                activation=nn.LeakyReLU(0.2, inplace=True),
                padding=1,
            ),
            *get_conv_blocks(
                out_channels,
                out_channels,
                3,
                1,
                norm_type=norm_type,
                activation=nn.LeakyReLU(0.2, inplace=True),
                padding=1,
            ),
        )

    def forward(self, net: Tensor, skip: Optional[Tensor] = None) -> Tensor:
        net = interpolate(net, mode="nearest", factor=2)
        if skip is not None:
            net = torch.cat([skip, interpolate(net, anchor=skip)], dim=1)
        return self.net(net)


class UNetDecoder(nn.Module):
    def __init__(self, backbone_channels: List[int], **kwargs: Any):
        super().__init__()
        in_channels = backbone_channels[::-1]
        skip_channels = in_channels[1:] + [0]
        out_channels = skip_channels.copy()
        out_channels[-1] = in_channels[-1]
        blocks = [
            UNetDecoderBlock(in_ch, skip_ch, out_ch, **kwargs)
            for in_ch, skip_ch, out_ch in zip(in_channels, skip_channels, out_channels)
        ]
        self.blocks = nn.ModuleList(blocks)

    def forward(self, backbone_outputs: tensor_dict_type) -> Tensor:
        indices = [int(k[5:]) for k in sorted(backbone_outputs)]
        features = [backbone_outputs[f"stage{i}"] for i in indices]
        features = features[::-1]

        net = features[0]
        skips = features[1:]
        for i, decoder_block in enumerate(self.blocks):
            skip = skips[i] if i < len(skips) else None
            net = decoder_block(net, skip)

        return net


@register_module("unet")
class UNet(nn.Module, ImageTranslatorMixin):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        backbone: str = "mobilenet_v2",
        backbone_config: Optional[Dict[str, Any]] = None,
        decoder_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.backbone = BackboneEncoder(
            backbone,
            in_channels,
            backbone_config=backbone_config,
        )
        backbone_channels = self.backbone.net.out_channels
        self.decoder = UNetDecoder(backbone_channels, **(decoder_config or {}))
        self.head = Conv2d(backbone_channels[0], out_channels, kernel_size=3)

    def forward(self, inp: Tensor) -> Tensor:
        features = self.backbone(inp)
        features.pop(LATENT_KEY)
        net = self.head(self.decoder(features))
        return {PREDICTIONS_KEY: interpolate(net, anchor=inp)}


__all__ = ["UNet"]



import torch

import torch.nn as nn
import torch.nn.functional as F

from abc import abstractmethod
from abc import ABC
from abc import ABCMeta
from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Optional
from torch.autograd import grad
from cftool.misc import shallow_copy_dict
from cftool.types import tensor_dict_type

from ..general import PureEncoderDecoder
from ..gan.discriminators import NLayerDiscriminator
from ...protocols import GaussianGeneratorMixin
from ....protocol import ITrainer
from ....protocol import TrainerState
from ....protocol import MetricsOutputs
from ....constants import INPUT_KEY
from ....constants import PREDICTIONS_KEY
from ....misc.internal_ import CustomModule
from ....misc.internal_ import CustomTrainStep
from ....misc.internal_ import CustomTrainStepLoss
from ....losses.lpips import LPIPS


class AutoEncoderInit(nn.Module):
    enc_double_channels: bool

    def __init__(
        self,
        img_size: int,
        in_channels: int,
        out_channels: int,
        inner_channels: int,
        latent_channels: int,
        channel_multipliers: Tuple[int, ...],
        *,
        embedding_channels: int,
        num_res_blocks: int,
        attention_resolutions: Tuple[int, ...] = (),
        dropout: float = 0.0,
        resample_with_conv: bool = True,
        attention_type: str = "spatial",
        apply_tanh: bool = False,
    ):
        super().__init__()
        self.img_size = img_size
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.inner_channels = inner_channels
        self.latent_channels = latent_channels
        self.channel_multipliers = channel_multipliers
        self.embedding_channels = embedding_channels
        module_config = dict(
            img_size=img_size,
            in_channels=in_channels,
            out_channels=out_channels,
            inner_channels=inner_channels,
            latent_channels=latent_channels,
            channel_multipliers=channel_multipliers,
            num_res_blocks=num_res_blocks,
            attention_resolutions=attention_resolutions,
            dropout=dropout,
            downsample_with_conv=resample_with_conv,
            upsample_with_conv=resample_with_conv,
            attention_type=attention_type,
        )
        enc_config, dec_config = map(shallow_copy_dict, 2 * [module_config])
        enc_scaler = 1 + int(self.enc_double_channels)
        enc_config["latent_channels"] = enc_scaler * latent_channels
        dec_config["apply_tanh"] = apply_tanh
        self.generator = PureEncoderDecoder(
            is_1d=False,
            encoder="attention",
            decoder="attention",
            encoder_config=enc_config,
            decoder_config=dec_config,
        )
        self.to_embedding = nn.Conv2d(
            enc_scaler * latent_channels,
            enc_scaler * embedding_channels,
            1,
        )
        self.from_embedding = nn.Conv2d(embedding_channels, latent_channels, 1)
        self.apply_tanh = apply_tanh


def d_hinge_loss(real: Tensor, fake: Tensor) -> Tensor:
    l_real = torch.mean(F.relu(1.0 - real))
    l_fake = torch.mean(F.relu(1.0 + fake))
    loss = 0.5 * (l_real + l_fake)
    return loss


def d_vanilla_loss(real: Tensor, fake: Tensor) -> Tensor:
    l_real = torch.mean(F.softplus(-real))
    l_fake = torch.mean(F.softplus(fake))
    loss = 0.5 * (l_real + l_fake)
    return loss


class AutoEncoderLPIPSWithDiscriminator(nn.Module, metaclass=ABCMeta):
    def __init__(
        self,
        *,
        kl_weight: float = 1.0,
        d_loss: str = "hinge",
        d_loss_start_step: int = 50001,
        d_num_layers: int = 4,
        d_in_channels: int = 3,
        d_start_channels: int = 64,
        d_factor: float = 1.0,
        d_weight: float = 1.0,
        perceptual_weight: float = 1.0,
    ):
        super().__init__()
        assert d_loss in ["hinge", "vanilla"]
        self.kl_weight = kl_weight
        self.perceptual_loss = LPIPS().eval()
        self.perceptual_weight = perceptual_weight
        self.discriminator = NLayerDiscriminator(
            in_channels=d_in_channels,
            num_layers=d_num_layers,
            start_channels=d_start_channels,
        )
        self.d_loss_start_step = d_loss_start_step
        self.d_loss = d_hinge_loss if d_loss == "hinge" else d_vanilla_loss
        self.d_factor = d_factor
        self.d_weight = d_weight

    def get_d_weight(
        self,
        nll_loss: Tensor,
        g_loss: Tensor,
        last_layer: nn.Parameter,
    ) -> Tensor:
        nll_grads = grad(nll_loss, last_layer, retain_graph=True)[0]
        g_grads = grad(g_loss, last_layer, retain_graph=True)[0]
        d_weight = torch.norm(nll_grads) / (torch.norm(g_grads) + 1.0e-4)
        d_weight = torch.clamp(d_weight, 0.0, 1.0e4).detach()
        d_weight = d_weight * self.d_weight
        return d_weight

    def g_loss(
        self,
        nll_loss: Tensor,
        last_layer: nn.Parameter,
        loss_items: Dict[str, float],
        reconstructions: Tensor,
        cond: Optional[Tensor],
    ) -> Tensor:
        device = nll_loss.device
        if cond is None:
            fake = self.discriminator(reconstructions).output
        else:
            fake = self.discriminator(torch.cat((reconstructions, cond), dim=1)).output
        g_loss = -torch.mean(fake)
        if self.d_factor <= 0.0:
            d_weight = torch.tensor(0.0, device=device)
        else:
            try:
                d_weight = self.get_d_weight(nll_loss, g_loss, last_layer=last_layer)
            except RuntimeError:
                assert not self.training
                d_weight = torch.tensor(0.0, device=device)
        loss_items["g_loss"] = g_loss.item()
        return d_weight * self.d_factor * g_loss

    def get_discriminator_loss(
        self,
        inputs: Tensor,
        reconstructions: Tensor,
        *,
        step: int,
        cond: Optional[Tensor] = None,
    ) -> CustomTrainStepLoss:
        if step < self.d_loss_start_step:
            raise ValueError(
                "should not call `get_discriminator_loss` because current step "
                f"({step}) is smaller than the `d_loss_start_step` "
                f"({self.d_loss_start_step})"
            )
        inputs = inputs.contiguous().detach()
        reconstructions = reconstructions.contiguous().detach()
        if cond is None:
            real = self.discriminator(inputs).output
            fake = self.discriminator(reconstructions).output
        else:
            real = self.discriminator(torch.cat((inputs, cond), dim=1)).output
            fake = self.discriminator(torch.cat((reconstructions, cond), dim=1)).output
        loss = self.d_factor * self.d_loss(real, fake)
        return CustomTrainStepLoss(loss, {"d_loss": loss.item()})

    @abstractmethod
    def get_generator_loss(
        self,
        batch: tensor_dict_type,
        forward_results: tensor_dict_type,
        *,
        step: int,
        last_layer: nn.Parameter,
        cond: Optional[Tensor] = None,
    ) -> CustomTrainStepLoss:
        pass


err_fmt = "`loss` is not initialized for `{}`"


class GeneratorStep(CustomTrainStep):
    def loss_fn(
        self,
        m: "AutoEncoderModelMixin",
        trainer: ITrainer,
        batch: tensor_dict_type,
        forward_results: tensor_dict_type,
        **kwargs: Any,
    ) -> CustomTrainStepLoss:
        if m.loss is None:
            raise ValueError(err_fmt.format(m.__class__.__name__))
        return m.loss.get_generator_loss(
            batch,
            forward_results,
            step=trainer.state.step,
            last_layer=m.generator.decoder.head[-1].weight,  # type: ignore
        )


class DiscriminatorStep(CustomTrainStep):
    def should_skip(self, m: "AutoEncoderModelMixin", state: TrainerState) -> bool:
        if m.loss is None:
            raise ValueError(err_fmt.format(m.__class__.__name__))
        return state.step < m.loss.d_loss_start_step

    def loss_fn(
        self,
        m: "AutoEncoderModelMixin",
        trainer: ITrainer,
        batch: tensor_dict_type,
        forward_results: tensor_dict_type,
        **kwargs: Any,
    ) -> CustomTrainStepLoss:
        if m.loss is None:
            raise ValueError(err_fmt.format(m.__class__.__name__))
        return m.loss.get_discriminator_loss(
            batch[INPUT_KEY],
            forward_results[PREDICTIONS_KEY],
            step=trainer.state.step,
        )


class AutoEncoderModelMixin(ABC):
    loss: Optional[AutoEncoderLPIPSWithDiscriminator]
    generator: nn.Module
    to_embedding: nn.Module
    from_embedding: nn.Module
    device: torch.device

    z_size: int
    grad_accumulate: int
    embedding_channels: int

    @property
    def use_loss(self) -> bool:
        return self.loss is not None

    @property
    def ae_parameters(self) -> List[nn.Parameter]:
        return (
            list(self.generator.parameters())
            + list(self.to_embedding.parameters())
            + list(self.from_embedding.parameters())
        )

    @property
    def d_parameters(self) -> List[nn.Parameter]:
        if self.loss is None:
            return []
        return list(self.loss.discriminator.parameters())

    @property
    def can_reconstruct(self) -> bool:
        return True

    @property
    def train_steps(self) -> List[CustomTrainStep]:
        g_scope = "core.ae_parameters"
        d_scope = "core.d_parameters"
        return [
            GeneratorStep(g_scope, grad_accumulate=self.grad_accumulate),
            DiscriminatorStep(
                d_scope,
                grad_accumulate=self.grad_accumulate,
                requires_new_forward=True,
                requires_grad_in_forward=True,
            ),
        ]

    def evaluate_step(
        self,
        batch: tensor_dict_type,
        state: TrainerState,
        trainer: ITrainer,
    ) -> MetricsOutputs:
        forward = self.forward(batch[INPUT_KEY])
        args = self, trainer, batch, forward
        loss_items = {}
        g_out = GeneratorStep().loss_fn(*args)
        loss_items.update(g_out.losses)
        if self.loss is None:
            raise ValueError(err_fmt.format(self.__class__.__name__))
        if state.step >= self.loss.d_loss_start_step:
            d_out = DiscriminatorStep().loss_fn(*args)
            loss_items.update(d_out.losses)
        score = -loss_items["recon"]
        return MetricsOutputs(score, loss_items)

    def generate_z(self, num_samples: int) -> Tensor:
        z = torch.randn(num_samples, self.embedding_channels, self.z_size, self.z_size)
        return z.to(self.device)

    def setup(
        self,
        img_size: int,
        grad_accumulate: int,
        embedding_channels: int,
        channel_multipliers: Tuple[int, ...],
    ) -> None:
        self.z_size = img_size // 2 ** len(channel_multipliers)
        self.embedding_channels = embedding_channels
        self.grad_accumulate = grad_accumulate

    @abstractmethod
    def forward(self, net: Tensor) -> tensor_dict_type:
        pass


class IAutoEncoder(  # type: ignore
    AutoEncoderModelMixin,
    CustomModule,
    GaussianGeneratorMixin,
    metaclass=ABCMeta,
):
    img_size: int


__all__ = [
    "AutoEncoderInit",
    "AutoEncoderModelMixin",
    "AutoEncoderLPIPSWithDiscriminator",
    "IAutoEncoder",
]
